import json
import os
from typing import Any, Callable, Iterable, List, Union

import torch
from torch.utils.data import DataLoader

import esp_ppq.lib as PFL
from esp_ppq.api.setting import DispatchingTable, QuantizationSetting, QuantizationSettingFactory
from esp_ppq.core import PPQ_CONFIG, NetworkFramework, TargetPlatform, empty_ppq_cache, ppq_warning
from esp_ppq.core.common import (
    FORMATTER_FORMAT_CONSTANT_INPUT,
    FORMATTER_FUSE_BIAS_ADD,
    FORMATTER_FUSE_BN,
    FORMATTER_FUSE_SWISH,
    FORMATTER_REMOVE_IDENTITY,
    FORMATTER_REMOVE_ISOLATED,
    FORMATTER_REPLACE_BN_TO_CONV,
)
from esp_ppq.executor import TorchExecutor
from esp_ppq.executor.base import BaseGraphExecutor
from esp_ppq.IR import (
    BaseGraph,
    GraphBuilder,
    GraphCommand,
    GraphCommandType,
    GraphFormatter,
    GraphMerger,
    GraphReplacer,
)
from esp_ppq.IR.quantize import QuantableOperation
from esp_ppq.quantization.optim.base import QuantizationOptimizationPass
from esp_ppq.quantization.quantizer import BaseQuantizer
from esp_ppq.scheduler import DISPATCHER_TABLE, GraphDispatcher


def load_graph(file_path: str, from_framework: NetworkFramework = NetworkFramework.ONNX, **kwargs) -> BaseGraph:
    parser = PFL.Parser(from_framework)
    assert isinstance(parser, GraphBuilder), 'Unexpected Parser found.'
    graph = parser.build(file_path)
    return graph


def load_onnx_graph(onnx_import_file: str) -> BaseGraph:
    """
        从一个指定位置加载 onnx 计算图，注意该加载的计算图尚未经过调度，此时所有算子被认为是可量化的
        load onnx graph from the specified location
    Args:
        onnx_import_file (str): onnx 计算图的保存位置 the specified location

    Returns:
        BaseGraph: 解析 onnx 获得的 ppq 计算图对象 the parsed ppq IR graph
    """
    ppq_ir = load_graph(onnx_import_file, from_framework=NetworkFramework.ONNX)
    return format_graph(graph=ppq_ir)


def load_native_graph(import_file: str) -> BaseGraph:
    """
        从一个指定位置加载 原生计算图，原生计算图中包含了所有量化与调度信息
        load native graph from the specified location
    Args:
        import_file (str): 计算图的保存位置 the specified location

    Returns:
        BaseGraph: 解析获得的 ppq 计算图对象 the parsed ppq IR graph
    """
    return load_graph(import_file, from_framework=NetworkFramework.NATIVE)


def load_torch_model(
    model: torch.nn.Module,
    onnx_file_name: str = 'MyModel.onnx',
    sample: Union[dict, list, torch.Tensor] = None,
    device='cuda',
    verbose: bool = False,
    opset: int = 11,
    do_constant_folding: bool = None,
    dynamic_axes: dict = None,
    training: bool = False,
) -> BaseGraph:
    """
    Convert a Pytorch Module to PPQ graph,
        this conversation exports pytorch model to onnx first,
        then PPQ will read from onnx file to build graph.

    Args:
        model (torch.nn.Module): Converting Model.

        onnx_file_name (str, optional): Exporting Onnx File name, by default it is 'MyModel.onnx'.

        sample (Union[dict, list, torch.Tensor], optional):
            A sample input for tracing your pytorch model.

            Pytorch will feed this sample input to your model,
            recording record every layer and functions during executing.

            Each function recorded will be translated into onnx.
            The input and output shape will also depends on this sample input.

        device (str, optional): _description_. Defaults to 'cuda'.

        verbose (bool, optional): _description_. Defaults to True.

        opset (int, optional): _description_. Defaults to 11.

        do_constant_folding (bool, optional):
            Apply the constant-folding optimization.
            Constant-folding will replace some of the ops that have all constant inputs
            with pre-computed constant nodes.

        dynamic_axes (dict, optional):
            (dict<string, dict<int, string>> or dict<string, list(int)>, default empty dict):
            By default the exported model will have the shapes of all input and output tensors
            set to exactly match those given in ``args``. To specify axes of tensors as
            dynamic (i.e. known only at run-time), set ``dynamic_axes`` to a dict with schema:

    Returns:
        BaseGraph: Converted PPQ Graph Instance.
    """

    if opset not in {11, 13}:
        ppq_warning(f'The Opset of Exporting Model should be 11 or 13, however {opset} was given.')

    model = model.eval()
    model = model.to(device)

    _training = torch.onnx.TrainingMode.TRAINING if training else torch.onnx.TrainingMode.EVAL
    torch.onnx.export(
        model=model,
        args=sample,
        verbose=verbose,
        f=onnx_file_name,
        opset_version=opset,
        do_constant_folding=do_constant_folding,
        dynamic_axes=dynamic_axes,
        training=_training,
    )

    return load_onnx_graph(onnx_import_file=onnx_file_name)


def dump_torch_to_onnx(
    model: torch.nn.Module,
    onnx_export_file: str,
    input_shape: List[int],
    input_dtype: torch.dtype = torch.float,
    inputs: List[Any] = None,
    device: str = 'cuda',
):
    """
        转换一个 torch 模型到 onnx，并保存到指定位置
        convert a torch model to onnx and save to the specified location
    Args:
        model (torch.nn.Module): 被转换的 torch 模型 torch model used for conversion

        onnx_export_file (str): 保存文件的路径 the path to save onnx model

        input_shape (List[int]): 模型输入尺寸，用于执行 jit.trace，对于动态尺寸的模型，输入一个模型可接受的尺寸即可。
            如果模型存在多个输入，则需要使用 inputs 变量进行传参，此项设置为 None
            a list of ints indicating size of input, for multiple inputs, please use keyword arg inputs for
            direct parameter passing and this should be set to None

        input_dtype (torch.dtype): 模型输入数据类型，如果模型存在多个输入，则需要使用 inputs 变量进行传参，此项设置为 None
                                   the torch datatype of input, for multiple inputs, please use keyword arg inputs
                                   for direct parameter passing and this should be set to None

        inputs (List[Any], optional): 对于存在多个输入的模型，在Inputs中直接指定一个输入List，从而完成模型的tracing。
                                    for multiple inputs, please give the specified inputs directly in the form of
                                    a list of arrays

        device (str, optional): 转换过程的执行设备 the execution device, defaults to 'cuda'.
    """

    # set model to eval mode, stabilize normalization weights.
    assert isinstance(model, torch.nn.Module), (
        f'Model must be instance of torch.nn.Module, however {type(model)} is given.'
    )
    model.eval()

    if inputs is None:
        dummy_input = torch.zeros(size=input_shape, device=device, dtype=input_dtype)
    else:
        dummy_input = inputs

    torch.onnx.export(
        model=model,
        args=dummy_input,
        verbose=False,
        f=onnx_export_file,
        opset_version=11,
    )


@empty_ppq_cache
def quantize_onnx_model(
    onnx_import_file: str,
    calib_dataloader: DataLoader,
    calib_steps: int,
    input_shape: List[int],
    platform: TargetPlatform,
    input_dtype: torch.dtype = torch.float,
    inputs: List[Any] = None,
    setting: QuantizationSetting = None,
    collate_fn: Callable = None,
    device: str = 'cuda',
    verbose: int = 0,
    do_quantize: bool = True,
) -> BaseGraph:
    """量化一个 onnx 原生的模型 输入一个 onnx 模型的文件路径 返回一个量化后的 esp_ppq.IR.BaseGraph quantize
    onnx model, input onnx model and return quantized ppq IR graph.

    Args:
        onnx_import_file (str): 被量化的 onnx 模型文件路径 onnx model location

        calib_dataloader (DataLoader): 校准数据集 calibration data loader

        calib_steps (int): 校准步数 calibration steps

        collate_fn (Callable): 校准数据的预处理函数 batch collate func for preprocessing

        input_shape (List[int]): 模型输入尺寸，用于执行 jit.trace，对于动态尺寸的模型，输入一个模型可接受的尺寸即可。
            如果模型存在多个输入，则需要使用 inputs 变量进行传参，此项设置为 None
                                a list of ints indicating size of input, for multiple inputs, please use
                                keyword arg inputs for direct parameter passing and this should be set to None

        input_dtype (torch.dtype): 模型输入数据类型，如果模型存在多个输入，则需要使用 inputs 变量进行传参，此项设置为 None
                                the torch datatype of input, for multiple inputs, please use keyword arg inputs
                                for direct parameter passing and this should be set to None

        inputs (List[Any], optional): 对于存在多个输入的模型，在Inputs中直接指定一个输入List，从而完成模型的tracing。
                                for multiple inputs, please give the specified inputs directly in the form of
                                a list of arrays

        setting (OptimSetting): 量化配置信息，用于配置量化的各项参数，设置为 None 时加载默认参数。
                                Quantization setting, default setting will be used when set None

        do_quantize (Bool, optional): 是否执行量化 whether to quantize the model, defaults to True.


        platform (TargetPlatform, optional): 量化的目标平台 target backend platform, defaults to TargetPlatform.DSP_INT8.

        device (str, optional): 量化过程的执行设备 execution device, defaults to 'cuda'.

        verbose (int, optional): 是否打印详细信息 whether to print details, defaults to 0.

    Raises:
        ValueError: 给定平台不可量化 the given platform doesn't support quantization
        KeyError: 给定平台不被支持 the given platform is not supported yet

    Returns:
        BaseGraph: 量化后的IR，包含了后端量化所需的全部信息
                   The quantized IR, containing all information needed for backend execution
    """
    if do_quantize:
        if calib_dataloader is None or calib_steps is None:
            raise TypeError('Quantization needs a valid calib_dataloader and calib_steps setting.')

    if setting is None:
        setting = QuantizationSettingFactory.default_setting()

    ppq_ir = load_onnx_graph(onnx_import_file=onnx_import_file)
    ppq_ir = dispatch_graph(
        graph=ppq_ir, platform=platform, dispatcher=setting.dispatcher, dispatching_table=setting.dispatching_table
    )

    if inputs is None:
        dummy_input = torch.zeros(size=input_shape, device=device, dtype=input_dtype)
    else:
        dummy_input = inputs

    quantizer = PFL.Quantizer(platform, ppq_ir)
    executor = TorchExecutor(graph=quantizer._graph, device=device)
    if do_quantize:
        quantizer.quantize(
            inputs=dummy_input,
            calib_dataloader=calib_dataloader,
            executor=executor,
            setting=setting,
            calib_steps=calib_steps,
            collate_fn=collate_fn,
        )
        if verbose:
            quantizer.report()
        return quantizer._graph
    else:
        executor = TorchExecutor(graph=ppq_ir, device=device)
        executor.tracing_operation_meta(inputs=dummy_input)
        return quantizer._graph


@empty_ppq_cache
def quantize_torch_model(
    model: torch.nn.Module,
    calib_dataloader: DataLoader,
    calib_steps: int,
    input_shape: List[int],
    platform: TargetPlatform,
    input_dtype: torch.dtype = torch.float,
    setting: QuantizationSetting = None,
    collate_fn: Callable = None,
    inputs: List[Any] = None,
    do_quantize: bool = True,
    onnx_export_file: str = 'onnx.model',
    device: str = 'cuda',
    verbose: int = 0,
) -> BaseGraph:
    """量化一个 Pytorch 原生的模型 输入一个 torch.nn.Module 返回一个量化后的 esp_ppq.IR.BaseGraph.

        quantize a pytorch model, input pytorch model and return quantized ppq IR graph
    Args:
        model (torch.nn.Module): 被量化的 torch 模型(torch.nn.Module) the pytorch model

        calib_dataloader (DataLoader): 校准数据集 calibration dataloader

        calib_steps (int): 校准步数 calibration steps

        collate_fn (Callable): 校准数据的预处理函数 batch collate func for preprocessing

        input_shape (List[int]): 模型输入尺寸，用于执行 jit.trace，对于动态尺寸的模型，输入一个模型可接受的尺寸即可。
            如果模型存在多个输入，则需要使用 inputs 变量进行传参，此项设置为 None
                                a list of ints indicating size of input, for multiple inputs, please use
                                keyword arg inputs for direct parameter passing and this should be set to None

        input_dtype (torch.dtype): 模型输入数据类型，如果模型存在多个输入，则需要使用 inputs 变量进行传参，此项设置为 None
                                the torch datatype of input, for multiple inputs, please use keyword arg inputs
                                for direct parameter passing and this should be set to None

        setting (OptimSetting): 量化配置信息，用于配置量化的各项参数，设置为 None 时加载默认参数。
                                Quantization setting, default setting will be used when set None

        inputs (List[Any], optional): 对于存在多个输入的模型，在Inputs中直接指定一个输入List，从而完成模型的tracing。
                                for multiple inputs, please give the specified inputs directly in the form of
                                a list of arrays

        do_quantize (Bool, optional): 是否执行量化 whether to quantize the model, defaults to True, defaults to True.

        platform (TargetPlatform, optional): 量化的目标平台 target backend platform, defaults to TargetPlatform.DSP_INT8.

        device (str, optional): 量化过程的执行设备 execution device, defaults to 'cuda'.

        verbose (int, optional): 是否打印详细信息 whether to print details, defaults to 0.

    Raises:
        ValueError: 给定平台不可量化 the given platform doesn't support quantization
        KeyError: 给定平台不被支持 the given platform is not supported yet

    Returns:
        BaseGraph: 量化后的IR，包含了后端量化所需的全部信息
                   The quantized IR, containing all information needed for backend execution
    """
    # dump pytorch model to onnx
    dump_torch_to_onnx(
        model=model,
        onnx_export_file=onnx_export_file,
        input_shape=input_shape,
        input_dtype=input_dtype,
        inputs=inputs,
        device=device,
    )

    return quantize_onnx_model(
        onnx_import_file=onnx_export_file,
        calib_dataloader=calib_dataloader,
        calib_steps=calib_steps,
        collate_fn=collate_fn,
        input_shape=input_shape,
        input_dtype=input_dtype,
        inputs=inputs,
        setting=setting,
        platform=platform,
        device=device,
        verbose=verbose,
        do_quantize=do_quantize,
    )


@empty_ppq_cache
def quantize_native_model(
    model: BaseGraph,
    calib_dataloader: DataLoader,
    calib_steps: int,
    input_shape: List[int],
    platform: TargetPlatform,
    input_dtype: torch.dtype = torch.float,
    inputs: List[Any] = None,
    setting: QuantizationSetting = None,
    collate_fn: Callable = None,
    device: str = 'cuda',
    verbose: int = 0,
    do_quantize: bool = True,
) -> BaseGraph:
    """量化一个已经在内存中的 ppq 模型 输入一个量化前的 esp_ppq.IR.BaseGraph 返回一个量化后的 esp_ppq.IR.BaseGraph
    quantize ppq model, input ppq graph and return quantized ppq graph.

    Args:
        native (BaseGraph): 被量化的 ppq graph

        calib_dataloader (DataLoader): 校准数据集 calibration data loader

        calib_steps (int): 校准步数 calibration steps

        collate_fn (Callable): 校准数据的预处理函数 batch collate func for preprocessing

        input_shape (List[int]): 模型输入尺寸，用于执行 jit.trace，对于动态尺寸的模型，输入一个模型可接受的尺寸即可。
            如果模型存在多个输入，则需要使用 inputs 变量进行传参，此项设置为 None
                                a list of ints indicating size of input, for multiple inputs, please use
                                keyword arg inputs for direct parameter passing and this should be set to None

        input_dtype (torch.dtype): 模型输入数据类型，如果模型存在多个输入，则需要使用 inputs 变量进行传参，此项设置为 None
                                the torch datatype of input, for multiple inputs, please use keyword arg inputs
                                for direct parameter passing and this should be set to None

        inputs (List[Any], optional): 对于存在多个输入的模型，在Inputs中直接指定一个输入List，从而完成模型的tracing。
                                for multiple inputs, please give the specified inputs directly in the form of
                                a list of arrays

        setting (OptimSetting): 量化配置信息，用于配置量化的各项参数，设置为 None 时加载默认参数。
                                Quantization setting, default setting will be used when set None

        do_quantize (Bool, optional): 是否执行量化 whether to quantize the model, defaults to True.


        platform (TargetPlatform, optional): 量化的目标平台 target backend platform, defaults to TargetPlatform.DSP_INT8.

        device (str, optional): 量化过程的执行设备 execution device, defaults to 'cuda'.

        verbose (int, optional): 是否打印详细信息 whether to print details, defaults to 0.

    Raises:
        ValueError: 给定平台不可量化 the given platform doesn't support quantization
        KeyError: 给定平台不被支持 the given platform is not supported yet

    Returns:
        BaseGraph: 量化后的IR，包含了后端量化所需的全部信息
                   The quantized IR, containing all information needed for backend execution
    """
    if do_quantize:
        if calib_dataloader is None or calib_steps is None:
            raise TypeError('Quantization needs a valid calib_dataloader and calib_steps setting.')

    if setting is None:
        setting = QuantizationSettingFactory.default_setting()
    ppq_ir = dispatch_graph(
        graph=model, platform=platform, dispatcher=setting.dispatcher, dispatching_table=setting.dispatching_table
    )

    if inputs is None:
        dummy_input = torch.zeros(size=input_shape, device=device, dtype=input_dtype)
    else:
        dummy_input = inputs

    quantizer = PFL.Quantizer(platform=platform, graph=ppq_ir)
    executor = TorchExecutor(graph=quantizer._graph, device=device)

    if do_quantize:
        quantizer.quantize(
            inputs=dummy_input,
            calib_dataloader=calib_dataloader,
            executor=executor,
            setting=setting,
            calib_steps=calib_steps,
            collate_fn=collate_fn,
        )
        if verbose:
            quantizer.report()
        return quantizer._graph
    else:
        executor = TorchExecutor(graph=ppq_ir, device=device)
        executor.tracing_operation_meta(inputs=dummy_input)
        return quantizer._graph


def export_ppq_graph(
    graph: BaseGraph,
    platform: TargetPlatform,
    graph_save_to: str,
    config_save_to: str = None,
    copy_graph: bool = True,
    **kwargs,
) -> None:
    """使用这个函数将 PPQ ir 保存到文件，同时导出 PPQ 的量化配置信息。 该函数可以将 PPQ ir 保存为不同格式的模型文件。 this
    func dumps ppq IR to file, and exports quantization setting information
    simultaneously.

    详细的支持情况请参考: esp_ppq.parser.__ini__.py
    for details please refer to esp_ppq.parser.__ini__.py

    Args:
        graph (BaseGraph): 被保存的 ir
                           the ppq IR graph

        platform (TargetPlatform): 期望部署的目标平台
                           target backend platform

        graph_save_to (str): 模型保存文件名，不要写后缀名，ppq 会自己加后缀
                           filename to save, do not add postfix to this

        config_save_to (str): 量化配置信息保存文件名。
            注意部分平台导出时会将量化配置信息直接写入模型，在这种情况下设置此参数无效
            note that some of platforms requires to write quantization setting
            directly into the model file, this parameter won't have effect at
            this situation

        copy_graph (bool): 导出图的时候是否需要把图复制一份
            Whether to copy graph when export.
    """
    for save_path in [graph_save_to, config_save_to]:
        if save_path is None:
            continue
        if os.path.exists(save_path):
            if os.path.isfile(save_path):
                ppq_warning(f'File {save_path} is already existed, Exporter will overwrite it.')
            if os.path.isdir(save_path):
                raise FileExistsError(
                    f'File {save_path} is already existed, and it is a directory, Exporter can not create file here.'
                )

    exporter = PFL.Exporter(platform)
    if copy_graph:
        graph = graph.copy()
    exporter.export(file_path=graph_save_to, config_path=config_save_to, graph=graph, **kwargs)


def format_graph(graph: BaseGraph) -> BaseGraph:
    """这个函数对计算图进行预处理工作，其主要内容是将计算图的格式进行统一 这个函数将会统一 cast, slice, parameter,
    constant 算子的格式，并且执行有关 batchnorm 的合并工作.

    在 PPQ 中，我们不希望出现 Constant 算子，所有 Constant 输入将被当作 parameter variable 连接到下游算子上
    在 PPQ 中，我们不希望出现 Batchnorm 算子，所有 Batchnorm 将被合并
    在 PPQ 中，我们不希望出现权重共享的算子，所有被共享的权重将被复制分裂成多份
    在 PPQ 中，我们不希望出现孤立算子，所有孤立算子将被移除

    This function takes pre-processing procedure with your graph.
    This function will convert operations like cast, slice, parameter, constant to the format that supported by esp_ppq.
    This function will merge batchnorm when possible.

    During quantization logic, we do not expect there is any constant operation in your network, so
        all of them will be converted as parameter input variable.

    We do not expect there is any shared parameter in your network, all of them will be copied and spilted.
    We do not expect any isolated operation in your network, all of them will be removed.
    """

    # do graph level optimization
    formatter = GraphReplacer(GraphFormatter(GraphMerger(graph)))
    if FORMATTER_FORMAT_CONSTANT_INPUT:
        formatter(GraphCommand(GraphCommandType.FORMAT_CONSTANT_INPUT))
    formatter(GraphCommand(GraphCommandType.CONVERT_TO_TENSOR))
    formatter(GraphCommand(GraphCommandType.FORMAT_PARAMETERS))

    if FORMATTER_FUSE_BIAS_ADD:
        formatter(GraphCommand(GraphCommandType.FUSE_BIAS_ADD))

    if FORMATTER_FUSE_BN:
        formatter(GraphCommand(GraphCommandType.FUSE_BN))

    if FORMATTER_REPLACE_BN_TO_CONV:
        formatter(GraphCommand(GraphCommandType.REPLACE_BATCHNORM_TO_CONV))

    if FORMATTER_FUSE_SWISH:
        formatter(GraphCommand(GraphCommandType.FUSE_SWISH))

    formatter(GraphCommand(GraphCommandType.FORMAT_CAST))
    formatter(GraphCommand(GraphCommandType.FORMAT_SLICE))
    formatter(GraphCommand(GraphCommandType.FORMAT_CLIP))
    formatter(GraphCommand(GraphCommandType.FORMAT_PAD))
    formatter(GraphCommand(GraphCommandType.FORMAT_RESIZE))

    if FORMATTER_REMOVE_IDENTITY:
        formatter(GraphCommand(GraphCommandType.REMOVE_IDENTITY))

    if FORMATTER_REMOVE_ISOLATED:
        formatter(GraphCommand(GraphCommandType.DELETE_ISOLATED))

    return graph


def dispatch_graph(
    graph: BaseGraph,
    platform: TargetPlatform,
    dispatcher: Union[str, GraphDispatcher] = 'conservative',
    dispatching_table: DispatchingTable = None,
) -> BaseGraph:
    """这个函数执行图切分与调度，你的计算图将被切分成一系列子图，并被调度到不同设备上。
    调度的逻辑分为自动控制的部分以及手动覆盖的部分，你可以使用 QuantizationSetting 来向这个函数传递手动调度表 从而覆盖 PPQ 的调度逻辑。

    注意：这个函数依据调度器和 TargetPlatform 平台的不同而产生行为差异，生成不同的调度计划。

    This function will cut your graph into a series of subgraph and send them to different device.
    PPQ provides an automatic dispatcher which, will generate different dispatching scheme on your TargetPlatform.
    A dispatching table can be passed via QuantizationSetting to override
        the default dispatching logic of ppq dispatcher manually.
    """
    dispatching_override = dispatching_table
    quantizer = PFL.Quantizer(platform=platform, graph=graph)  # 初始化一个 quantizer 没有很大代价...

    if isinstance(dispatcher, str):
        dispatcher = dispatcher.lower()
        if dispatcher not in DISPATCHER_TABLE:
            raise ValueError(f'Can not found dispatcher type "{dispatcher}", check your input again.')
        dispatcher = DISPATCHER_TABLE[dispatcher](graph)
    else:
        if not isinstance(dispatcher, GraphDispatcher):
            raise TypeError(
                'Parameter "dispachter" of function esp_ppq.api.dispatch_graph must be String or GraphDispatcher, '
                f'however {type(dispatcher)} was given.'
            )
        dispatcher = dispatcher

    assert isinstance(dispatcher, GraphDispatcher)
    assert isinstance(quantizer, BaseQuantizer)
    quant_types = quantizer.quant_operation_types
    dispatching_table = dispatcher.dispatch(
        graph=graph,
        quant_types=quant_types,
        quant_platform=TargetPlatform.UNSPECIFIED,  # MUST BE UNSPECIFIED, 这里的意思是交由 Quantizer 决定是否量化这个算子
        fp32_platform=TargetPlatform.FP32,
        SOI_platform=TargetPlatform.SOI,
    )

    # override dispatching result
    if dispatching_override is not None:
        if not isinstance(dispatching_override, DispatchingTable):
            raise TypeError(
                'Parameter "dispatching_table" of function esp_ppq.api.dispatch_graph must be DispatchingTable, '
                f'however {type(dispatching_override)} was given.'
            )

        for opname, platform in dispatching_override.dispatchings.items():
            if opname not in graph.operations:
                continue
            assert isinstance(platform, int), (
                f'Your dispatching table contains a invalid setting of operation {opname}, '
                'All platform setting given in dispatching table is expected given as int, '
                f'however {type(platform)} was given.'
            )
            dispatching_table[opname] = TargetPlatform(platform)

    for operation in graph.operations.values():
        assert operation.name in dispatching_table, (
            f'Internal Error, Can not find operation {operation.name} in dispatching table.'
        )
        operation.platform = dispatching_table[operation.name]
    return graph


class UnbelievableUserFriendlyQuantizationSetting:
    """
    量化配置文件 -- 入门版

    这个文件包含了最基本的量化配置。
    """

    def __init__(
        self,
        platform: TargetPlatform,
        finetune_steps: int = 500,
        finetune_lr: float = 3e-5,
        interested_outputs: List[str] = None,
        calibration: str = 'percentile',
        equalization: bool = True,
        non_quantable_op: List[str] = None,
    ) -> None:
        """
        量化配置文件 -- 入门版

        这个文件包含了最基本的量化配置。

        Args:
            platform (TargetPlatform): 目标量化平台
            finetune_steps (int, optional): 网络 finetune 步数. Defaults to 5000.
            finetune_lr (float, optional): 网络 finetune 学习率. Defaults to 3e-4.
            interested_outputs (List[str], optional): 用来finetune的variable名字，请注意对于静态图而言其总是由 op 和 variable 组成的，
                有时候你的网络输出并不是可导的，或者是一个softmax或者sigmoid的输出，这些时候finetune的结果不会很好，你可以通过这个属性
                来指定一个variable的名字，我们将用这个variable的输出结果来引导finetune流程，当然一个variable list也是可以的。
            equalization (bool, optional): 是否要拉平网络权重. Defaults to True.
            non_quantable_op (List[str], optional): 非量化算子集合，所有名字出现在该集合里的算子将不被量化. Defaults to None.
        """
        self.equalization = equalization
        self.finetune_steps = finetune_steps
        self.finetune_lr = finetune_lr
        self.calibration = calibration
        self.platform = platform
        self.non_quantable_op = non_quantable_op
        self.interested_outputs = interested_outputs

        if isinstance(self.non_quantable_op, str):
            self.non_quantable_op = [self.non_quantable_op]
        if isinstance(self.interested_outputs, str):
            self.interested_outputs = [self.interested_outputs]

    def convert_to_daddy_setting(self) -> QuantizationSetting:
        # 将菜鸡版量化配置转换成高级版的
        daddy = QuantizationSettingFactory.default_setting()
        daddy.quantize_activation_setting.calib_algorithm = self.calibration

        if self.platform in {TargetPlatform.METAX_INT8_C, TargetPlatform.METAX_INT8_T}:
            daddy.fusion_setting.force_alignment_overlap = True

        if self.finetune_steps > 0:
            daddy.lsq_optimization = True
            daddy.lsq_optimization_setting.steps = self.finetune_steps
            daddy.lsq_optimization_setting.lr = self.finetune_lr

        if self.equalization == True:
            daddy.equalization = True
            daddy.equalization_setting.iterations = 3
            daddy.equalization_setting.opt_level = 1
            daddy.equalization_setting.value_threshold = 0

        if self.non_quantable_op is not None:
            for op_name in self.non_quantable_op:
                assert isinstance(op_name, str), (
                    f'你尝试使用 non_quantable_op 来设定非量化算子，'
                    f'non_quantable_op 只应当包含算子的名字，而你传入的数据中包括了 {type(op_name)}'
                )
                daddy.dispatching_table.append(op_name, TargetPlatform.FP32)

        return daddy

    def to_json(self, file_path: str) -> str:
        if os.path.exists(file_path):
            if os.path.isdir(file_path):
                raise FileExistsError(f'文件 {file_path} 已经存在且是一个目录，无法将配置文件写入到该位置！')
            ppq_warning(f'文件 {file_path} 已经存在并将被覆盖')

        # TargetPlatform is not a native type, convert it to string.
        dump_dict = self.__dict__.copy()
        dump_dict['platform'] = self.platform.name

        with open(file_path, 'w', encoding='utf-8') as file:
            json.dump(obj=dump_dict, fp=file, sort_keys=True, indent=4, ensure_ascii=False)

    @staticmethod
    def from_file(file_path: str):
        if not os.path.exists(file_path):
            raise FileNotFoundError('找不到你的配置文件，检查配置文件路径是否正确！')
        with open(file_path, 'r', encoding='utf-8') as file:
            loaded = json.load(file)
        assert isinstance(loaded, dict), 'Json文件无法解析，格式不正确'
        assert 'platform' in loaded, 'Json文件缺少必要项目 "platform"'

        platform = loaded['platform']
        if platform in TargetPlatform._member_names_:
            platform = TargetPlatform._member_map_[platform]
        else:
            raise KeyError('无法解析你的json配置文件，遇到了未知的platform属性。')

        setting = UnbelievableUserFriendlyQuantizationSetting(platform)
        for key, value in loaded.items():
            if key == 'platform':
                continue
            if key in setting.__dict__:
                setting.__dict__[key] = value
            if key not in setting.__dict__:
                ppq_warning(f'你的Json文件中包含无法解析的属性 {key} ，该属性已经被舍弃')
        assert isinstance(setting, UnbelievableUserFriendlyQuantizationSetting)
        return setting

    def __str__(self) -> str:
        return str(self.__dict__)


def quantize(
    working_directory: str,
    setting: QuantizationSetting,
    model_type: NetworkFramework,
    executing_device: str,
    input_shape: List[int],
    target_platform: TargetPlatform,
    dataloader: DataLoader,
    calib_steps: int = 32,
) -> BaseGraph:
    """Helper function for quantize your model within working directory, This
    function will do some check and redirect your requirement to:
    esp_ppq.api.quantize_onnx_model.

    see them for more information.

    Args:
        working_directory (str): A path that indicates working directory.
        setting (QuantizationSetting): Quantization setting
        model_type (NetworkFramework): Onnx
        executing_device (str): 'cuda' or 'cpu'
        input_shape (List[int]): sample input's shape
        target_platform (TargetPlatform): Target deploy platform
        dataloader (DataLoader): calibraiton dataloader
        calib_steps (int, optional): Defaults to 32.

    Raises:
        FileNotFoundError: _description_
        FileNotFoundError: _description_

    Returns:
        BaseGraph: _description_
    """
    if model_type == NetworkFramework.ONNX:
        if not os.path.exists(os.path.join(working_directory, 'model.onnx')):
            raise FileNotFoundError(f'无法找到你的模型: {os.path.join(working_directory, "model.onnx")}')
        return quantize_onnx_model(
            onnx_import_file=os.path.join(working_directory, 'model.onnx'),
            calib_dataloader=dataloader,
            calib_steps=calib_steps,
            input_shape=input_shape,
            setting=setting,
            platform=target_platform,
            device=executing_device,
            collate_fn=lambda x: x.to(executing_device),
        )


def export(working_directory: str, quantized: BaseGraph, platform: TargetPlatform, **kwargs):
    """Helper function to export your graph to working directory, You should
    notice this function just redirect your invoking to export_ppq_graph. see
    export_ppq_graph for more information.

    Args:
        working_directory (str): _description_
        quantized (BaseGraph): _description_
        platform (TargetPlatform): _description_
    """
    export_ppq_graph(
        graph=quantized,
        platform=platform,
        graph_save_to=os.path.join(working_directory, 'quantized'),
        config_save_to=os.path.join(working_directory, 'quantized.json'),
        **kwargs,
    )


def manop(
    graph: BaseGraph,
    list_of_passes: List[QuantizationOptimizationPass],
    calib_dataloader: Iterable,
    executor: BaseGraphExecutor,
    collate_fn: Callable = None,
    **kwargs,
) -> BaseGraph:
    """manop 是一个很方便的函数，你可以调用这个函数来手动地执行一些量化优化工作
    你可以在默认量化逻辑之前或之后调用这个函数来自定义量化处理流程，相比于直接实现 Quantizer来修改量化逻辑的方式, 使用 manop
    会更加灵活。

    MANOP (manually optimize) function is introduced since PPQ 0.6.4.
    This function allows you to manually invoke
        QuantizationOptimizationPass before or after default quantization logic.

    We do not use function name like apply, optim, do ...
    Because they are so widely-used in other python libraries,
        and can easily conflict with each other.

    Args:
        graph (BaseGraph): processing graph.
        list_of_passes (List[QuantizationOptimizationPass]): a collection of optimization logic.
        calib_dataloader (Iterable): _description_
        executor (BaseGraphExecutor): _description_
        collate_fn (Callable): _description_

    Raises:
        TypeError: _description_
        TypeError: _description_

    Returns:
        BaseGraph: processed graph
    """
    if isinstance(list_of_passes, QuantizationOptimizationPass):
        list_of_passes = [list_of_passes]

    if not (isinstance(list_of_passes, list) or isinstance(list_of_passes, tuple)):
        raise TypeError(
            'Can not apply optimization on your graph, '
            'expect a list of QuantizationOptimizationPass as input, '
            f'while {type(list_of_passes)} was given.'
        )

    for optim in list_of_passes:
        if not isinstance(optim, QuantizationOptimizationPass):
            raise TypeError(
                'Invoking this function needs a list of QuantizationOptimizationPass, '
                f'however there is a/an {type(optim)} in your list'
            )
        optim.apply(graph, dataloader=calib_dataloader, executor=executor, collate_fn=collate_fn, **kwargs)
    return graph


class ENABLE_CUDA_KERNEL:
    """Any code surrounded by
    with ENABLE_CUDA_KERNEL():
    will invoke ppq's kernel functions for speed boost.

    This is a helper class for invoking highly-effcient custimized cuda
    kernel. PPQ developer team has implemented a series of quantization related
    cuda kernel, They are 5-100x faster than torch kernels, with less gpu
    memory cost.
    """

    def __init__(self) -> None:
        from esp_ppq.core.ffi import CUDA_COMPLIER

        CUDA_COMPLIER.complie()
        self._state = False

    def __enter__(self):
        self._state = PPQ_CONFIG.USING_CUDA_KERNEL
        PPQ_CONFIG.USING_CUDA_KERNEL = True

    def __exit__(self, *args):
        PPQ_CONFIG.USING_CUDA_KERNEL = self._state


class DISABLE_CUDA_KERNEL:
    """Any code surrounded by
    with DISABLE_CUDA_KERNEL():
    will block ppq's kernel functions.

    This is a helper class for blocking ppq custimized cuda
    kernel. PPQ developer team has implemented a series of quantization related
    cuda kernel, They are 5-100x faster than torch kernels, with less gpu
    memory cost.
    """

    def __init__(self) -> None:
        pass

    def __enter__(self):
        self._state = PPQ_CONFIG.USING_CUDA_KERNEL
        PPQ_CONFIG.USING_CUDA_KERNEL = False

    def __exit__(self, *args):
        PPQ_CONFIG.USING_CUDA_KERNEL = self._state


class DEQUANTIZE_GRAPH:
    """A short-cut function to dequantize given graph."""

    def __init__(self, graph: BaseGraph) -> None:
        self._graph = graph
        self._state_recorder = {}
        self._quantable_ops = [op for op in graph.operations.values() if isinstance(op, QuantableOperation)]

    def __enter__(self):
        for op in self._quantable_ops:
            op.dequantize()

    def __exit__(self, *args):
        for op in self._quantable_ops:
            op.restore_quantize_state()


__all__ = [
    'load_graph',
    'load_onnx_graph',
    'dispatch_graph',
    'dump_torch_to_onnx',
    'quantize_onnx_model',
    'quantize_torch_model',
    'load_native_graph',
    'export_ppq_graph',
    'format_graph',
    'quantize',
    'export',
    'load_torch_model',
    'UnbelievableUserFriendlyQuantizationSetting',
    'manop',
    'quantize_native_model',
    'ENABLE_CUDA_KERNEL',
    'DISABLE_CUDA_KERNEL',
    'DEQUANTIZE_GRAPH',
]



from esp_ppq.core import NetworkFramework, TargetPlatform
from esp_ppq.parser import (
    AscendExporter,
    EspdlExporter,
    ExtensionExporter,
    MNNExporter,
    NativeExporter,
    NativeImporter,
    NCNNExporter,
    NxpExporter,
    OnnxExporter,
    OnnxParser,
    ONNXRUNTIMExporter,
    OpenvinoExporter,
    PPLBackendExporter,
    QNNDSPExporter,
    TengineExporter,
    TensorRTExporter_JSON,
)
from esp_ppq.quantization.quantizer import (
    AscendQuantizer,
    EspdlCHPreInt16Quantizer,
    EspdlCInt16Quantizer,
    EspdlCQuantizer,
    EspdlHPreInt16Quantizer,
    EspdlInt16Quantizer,
    EspdlQuantizer,
    EspdlS3HPreInt16Quantizer,
    EspdlS3Int16Quantizer,
    EspdlS3Quantizer,
    ExtQuantizer,
    FPGAQuantizer,
    GraphCoreQuantizer,
    MetaxChannelwiseQuantizer,
    MetaxTensorwiseQuantizer,
    MNNQuantizer,
    NCNNQuantizer,
    NXP_Quantizer,
    OnnxruntimeQuantizer,
    OpenvinoQuantizer,
    PPL_DSP_Quantizer,
    PPL_DSP_TI_Quantizer,
    PPLCUDAQuantizer,
    RKNN_PerTensorQuantizer,
    TengineQuantizer,
    TensorRTQuantizer,
    TensorRTQuantizer_FP8,
)

__QUANTIZER_COLLECTION__ = {
    TargetPlatform.PPL_DSP_INT8: PPL_DSP_Quantizer,
    TargetPlatform.PPL_DSP_TI_INT8: PPL_DSP_TI_Quantizer,
    TargetPlatform.SNPE_INT8: PPL_DSP_Quantizer,
    TargetPlatform.QNN_DSP_INT8: PPL_DSP_Quantizer,
    TargetPlatform.TRT_INT8: TensorRTQuantizer,
    TargetPlatform.MNN_INT8: MNNQuantizer,
    TargetPlatform.ASC_INT8: AscendQuantizer,
    TargetPlatform.NCNN_INT8: NCNNQuantizer,
    TargetPlatform.NXP_INT8: NXP_Quantizer,
    TargetPlatform.RKNN_INT8: RKNN_PerTensorQuantizer,
    TargetPlatform.METAX_INT8_C: MetaxChannelwiseQuantizer,
    TargetPlatform.METAX_INT8_T: MetaxTensorwiseQuantizer,
    TargetPlatform.PPL_CUDA_INT8: PPLCUDAQuantizer,
    TargetPlatform.EXTENSION: ExtQuantizer,
    TargetPlatform.FPGA_INT8: FPGAQuantizer,
    TargetPlatform.OPENVINO_INT8: OpenvinoQuantizer,
    TargetPlatform.TENGINE_INT8: TengineQuantizer,
    TargetPlatform.GRAPHCORE_FP8: GraphCoreQuantizer,
    TargetPlatform.TRT_FP8: TensorRTQuantizer_FP8,
    TargetPlatform.ONNXRUNTIME: OnnxruntimeQuantizer,
    TargetPlatform.ESPDL_INT8: EspdlQuantizer,
    TargetPlatform.ESPDL_INT16: EspdlInt16Quantizer,
    TargetPlatform.ESPDL_S3_INT8: EspdlS3Quantizer,
    TargetPlatform.ESPDL_S3_INT16: EspdlS3Int16Quantizer,
    TargetPlatform.ESPDL_H_PRE_INT16: EspdlHPreInt16Quantizer,
    TargetPlatform.ESPDL_S3_H_PRE_INT16: EspdlS3HPreInt16Quantizer,
    TargetPlatform.ESPDL_C_INT8: EspdlCQuantizer,
    TargetPlatform.ESPDL_C_INT16: EspdlCInt16Quantizer,
    TargetPlatform.ESPDL_C_H_PRE_INT16: EspdlCHPreInt16Quantizer,
}


__PARSERS__ = {NetworkFramework.ONNX: OnnxParser, NetworkFramework.NATIVE: NativeImporter}


__EXPORTERS__ = {
    # TargetPlatform.PPL_DSP_INT8:  PPLDSPCaffeExporter,
    # TargetPlatform.PPL_DSP_TI_INT8: PPLDSPTICaffeExporter,
    TargetPlatform.QNN_DSP_INT8: QNNDSPExporter,
    TargetPlatform.PPL_CUDA_INT8: PPLBackendExporter,
    # TargetPlatform.SNPE_INT8:     SNPECaffeExporter,
    TargetPlatform.NXP_INT8: NxpExporter,
    TargetPlatform.ONNX: OnnxExporter,
    TargetPlatform.ONNXRUNTIME: ONNXRUNTIMExporter,
    TargetPlatform.OPENVINO_INT8: OpenvinoExporter,
    # TargetPlatform.CAFFE:         CaffeExporter,
    TargetPlatform.NATIVE: NativeExporter,
    TargetPlatform.EXTENSION: ExtensionExporter,
    TargetPlatform.RKNN_INT8: OnnxExporter,
    TargetPlatform.METAX_INT8_C: ONNXRUNTIMExporter,
    TargetPlatform.METAX_INT8_T: ONNXRUNTIMExporter,
    TargetPlatform.TRT_INT8: TensorRTExporter_JSON,
    TargetPlatform.ASC_INT8: AscendExporter,
    TargetPlatform.TRT_FP8: ONNXRUNTIMExporter,
    TargetPlatform.NCNN_INT8: NCNNExporter,
    TargetPlatform.TENGINE_INT8: TengineExporter,
    TargetPlatform.MNN_INT8: MNNExporter,
    TargetPlatform.ESPDL_INT8: EspdlExporter,
    TargetPlatform.ESPDL_INT16: EspdlExporter,
    TargetPlatform.ESPDL_S3_INT8: EspdlExporter,
    TargetPlatform.ESPDL_S3_INT16: EspdlExporter,
    TargetPlatform.ESPDL_H_PRE_INT16: EspdlExporter,
    TargetPlatform.ESPDL_S3_H_PRE_INT16: EspdlExporter,
    TargetPlatform.ESPDL_C_INT8: EspdlExporter,
    TargetPlatform.ESPDL_C_INT16: EspdlExporter,
    TargetPlatform.ESPDL_C_H_PRE_INT16: EspdlExporter,
}


__all__ = ['__QUANTIZER_COLLECTION__', '__PARSERS__', '__EXPORTERS__']



# Test Quantization System Performance on Detection Models with Coco Dataset

# Quantizer Configuration
SYMMETRICAL = True
PER_CHANNEL = False
POWER_OF_2 = True
BIT_WIDTH = 8
FP8 = False

# Should contains model file(.onnx)
MODEL_DIR = 'QuantZoo/Model/yolo'

# Should contains Calib & Test Img Folder
CALIB_DIR = 'QuantZoo/Data/Coco/Calib'
TEST_DIR = 'QuantZoo/Data/Coco/Test'
CALIB_ANN_FILE = 'QuantZoo/Data/Coco/Calib/DetectionAnnotation.json'
TEST_ANN_FILE = 'QuantZoo/Data/Coco/Test/DetectionAnnotation.json'
PRED_ANN_FILE = 'QuantZoo/Data/Coco/Test/DetectionPrediction.json'
VALID_DIR = '/mnt/hpc/share/wangpeiqi/COCO/val2017'
VALID_ANN_FILE = '/mnt/hpc/share/wangpeiqi/COCO/annotations/instances_val2017.json'
EVAL_MODE = True  # only for evaluation, it will slow down the system.

# calibration & test batchsize
# yolo requires batchsize = 1
BATCHSIZE = 1

# write report to here
REPORT_DIR = 'QuantZoo/Reports'

CONFIGS = [
    {
        'Model': 'yolov6p5_n',
        'Output': ['/Concat_5_output_0', '/Concat_4_output_0'],
        'collate_fn': lambda x: x[0].cuda(),  # img preprocessing function
    },
    {
        'Model': 'yolov6p5_t',
        'Output': ['/Concat_5_output_0', '/Concat_4_output_0'],
        'collate_fn': lambda x: x[0].cuda(),  # img preprocessing function
    },
    {
        'Model': 'yolov5s6_n',
        'Output': [
            '/baseModel/head_module/convs_pred.1/Conv_output_0',
            '/baseModel/head_module/convs_pred.2/Conv_output_0',
            '/baseModel/head_module/convs_pred.0/Conv_output_0',
        ],
        'collate_fn': lambda x: x[0].cuda(),  # img preprocessing function
    },
    {
        'Model': 'yolov5s6_s',
        'Output': [
            '/baseModel/head_module/convs_pred.1/Conv_output_0',
            '/baseModel/head_module/convs_pred.2/Conv_output_0',
            '/baseModel/head_module/convs_pred.0/Conv_output_0',
        ],
        'collate_fn': lambda x: x[0].cuda(),  # img preprocessing function
    },
    {
        'Model': 'yolov7p5_tiny',
        'Output': ['/Concat_4_output_0', '/Concat_5_output_0', '/Concat_6_output_0'],
        'collate_fn': lambda x: x[0].cuda(),  # img preprocessing function
    },
    {
        'Model': 'yolov7p5_l',
        'Output': ['/Concat_4_output_0', '/Concat_5_output_0', '/Concat_6_output_0'],
        'collate_fn': lambda x: x[0].cuda(),  # img preprocessing function
    },
    {
        'Model': 'yolox_s',
        'Output': ['/Concat_4_output_0', '/Concat_5_output_0', '/Concat_6_output_0'],
        'collate_fn': lambda x: x[0].cuda() * 255,  # img preprocessing function
    },
    {
        'Model': 'yolox_tiny',
        'Output': ['/Concat_4_output_0', '/Concat_5_output_0', '/Concat_6_output_0'],
        'collate_fn': lambda x: x[0].cuda() * 255,  # img preprocessing function
    },
    {
        'Model': 'ppyoloe_m',
        'Output': ['/Concat_4_output_0', '/Concat_5_output_0'],
        'collate_fn': lambda x: (
            x[0].cuda() * 255 - torch.tensor([103.53, 116.28, 123.675]).reshape([1, 3, 1, 1]).cuda()
        )
        / 255,  # img preprocessing function
    },
    {
        'Model': 'ppyoloe_s',
        'Output': ['/Concat_4_output_0', '/Concat_5_output_0'],
        'collate_fn': lambda x: (
            x[0].cuda() * 255 - torch.tensor([103.53, 116.28, 123.675]).reshape([1, 3, 1, 1]).cuda()
        )
        / 255,  # img preprocessing function
    },
]

import os

import torch
from QuantZoo.Data.Coco.Data import load_coco_detection_dataset
from QuantZoo.Data.Coco.Eval import evaluate_ppq_module_with_coco
from QuantZoo.Quantizers import MyFP8Quantizer, MyInt8Quantizer
from QuantZoo.Util import error_analyze

import esp_ppq.lib as PFL
from esp_ppq.api import ENABLE_CUDA_KERNEL, load_onnx_graph
from esp_ppq.core import TargetPlatform
from esp_ppq.executor import TorchExecutor
from esp_ppq.IR import GraphFormatter
from esp_ppq.quantization.optim import (
    LayerwiseEqualizationPass,
    LearnedStepSizePass,
    ParameterQuantizePass,
    RuntimeCalibrationPass,
)

calib_loader = load_coco_detection_dataset(data_dir=CALIB_DIR, batchsize=BATCHSIZE)

test_loader = load_coco_detection_dataset(data_dir=TEST_DIR, batchsize=BATCHSIZE)


with ENABLE_CUDA_KERNEL():
    for config in CONFIGS:
        model = config['Model']
        monitoring_vars = config['Output']
        collate_fn = config['collate_fn']

        print(f"Ready to run quant benchmark on {model}")
        graph = load_onnx_graph(onnx_import_file=os.path.join(MODEL_DIR, model + '.onnx'))

        # if EVAL_MODE == False, truncate graph
        if EVAL_MODE == False:
            graph.outputs.clear()
            editor = GraphFormatter(graph)
            for var in monitoring_vars:
                graph.mark_variable_as_graph_output(graph.variables[var])
            editor.delete_isolated()
        else:
            if model in {'rtmdet_s', 'rtmdet_tiny'}:
                # 别问我为什么，只是因为大家的输出都叫 '/Split_output_1'，但是 rtmdet 的不叫这个
                # rename variable 'onnx::Shape_1182' to '/Split_output_1'
                graph.variables['/Split_output_1'] = graph.variables['onnx::Shape_1182']
                graph.variables['/Split_output_1']._name = '/Split_output_1'

            editor = GraphFormatter(graph)
            graph.outputs.pop('scores')
            graph.outputs.pop('num_dets')
            graph.mark_variable_as_graph_output(graph.variables['/Split_output_1'])
            editor.delete_isolated()

        quantizer = MyInt8Quantizer(
            graph=graph, sym=SYMMETRICAL, per_channel=PER_CHANNEL, power_of_2=POWER_OF_2, num_of_bits=BIT_WIDTH
        )
        if FP8:
            quantizer = MyFP8Quantizer(graph=graph, calibration='floating')

        # convert op to quantable-op
        for name, op in graph.operations.items():
            if op.type in {'Conv', 'ConvTranspose', 'MatMul', 'Gemm'}:
                quantizer.quantize_operation(name, platform=TargetPlatform.INT8)

        # build quant pipeline.
        pipeline = PFL.Pipeline(
            [
                # LayerwiseEqualizationPass(iteration=10),
                ParameterQuantizePass(),
                RuntimeCalibrationPass(),
                # LearnedStepSizePass(steps=500, collecting_device='cpu')
            ]
        )

        # call pipeline.
        executor = TorchExecutor(graph=graph)
        executor.tracing_operation_meta(torch.zeros(size=[BATCHSIZE, 3, 640, 640]).cuda())

        pipeline.optimize(
            graph=graph, dataloader=calib_loader, verbose=True, calib_steps=32, collate_fn=collate_fn, executor=executor
        )

        if EVAL_MODE:
            # evaluation 好像 batchsize != 1 会错
            evaluate_ppq_module_with_coco(
                ann_file=TEST_ANN_FILE,
                output_file=PRED_ANN_FILE,
                executor=executor,
                dataloader=test_loader,
                collate_fn=collate_fn,
            )

        # error analyze
        performance = error_analyze(
            graph=graph, outputs=monitoring_vars, dataloader=test_loader, collate_fn=collate_fn, verbose=True
        )



from abc import abstractmethod
from collections import deque
from typing import Any, Dict, List, Text, Union

import numpy as np
import torch

from esp_ppq.core import (
    LINEAR_ACTIVATIONS,
    DataType,
    NetworkFramework,
    Serializable,
    SingletonMeta,
    TargetPlatform,
    TensorMeta,
    convert_any_to_torch_tensor,
    ppq_warning,
)

from .opdef import DEFAULT_SOCKET_CREATOR, DEFAULT_SOCKET_TABLE, OperationBase, Opset, OpSocket


class Variable(Serializable):
    def __init__(
        self,
        name: str,
        value: Any = None,
        is_parameter: bool = False,
        dest_ops: List[OperationBase] = None,
        source_op: OperationBase = None,
        shape: List[int] = None,
        dtype: DataType = DataType.FP32,
    ) -> None:
        super().__init__()
        self._name = name
        self._value = None if value is None else value
        self._dest_ops = [] if dest_ops is None else dest_ops
        self._source_op = None if source_op is None else source_op
        self._is_parameter = is_parameter
        self.shape = shape
        self.dtype = dtype

    @property
    def is_parameter(self) -> bool:
        return self._is_parameter

    @is_parameter.setter
    def is_parameter(self, value: bool):
        self._is_parameter = value

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> torch.Tensor:
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @property
    def dest_ops(self) -> List[OperationBase]:
        return self._dest_ops

    @property
    def source_op(self) -> OperationBase:
        return self._source_op

    @source_op.setter
    def source_op(self, source_op: OperationBase) -> OperationBase:
        self._source_op = source_op

    @property
    def dest_idx(self) -> List[int]:
        _dest_idx = []
        for op in self.dest_ops:
            _dest_idx.append(op.inputs.index(self))
        return _dest_idx

    @property
    def src_idx(self) -> int:
        if self.source_op is not None:
            return self.source_op.outputs.index(self)
        else:
            return None

    @property
    def meta(self) -> TensorMeta:
        raise Exception('PPQ Variable.meta has been removed since 0.6.6, use Variable.shape, Variable.dtype instead.')

    def __hash__(self) -> int:
        return self._name.__hash__()

    def __str__(self) -> str:
        return f'{self._name}'

    def __repr__(self) -> str:
        if self.shape is not None:
            return f'{self._name}(shape={self.shape})'
        return f'{self._name}'

    def __getstate__(self) -> dict:
        state = super().__getstate__()
        state['_dest_ops'] = [op.name for op in self.dest_ops]
        state['_source_op'] = self.source_op.name if self.source_op is not None else None
        return state

    @property
    def shape(self) -> List[Union[Text, int, None]]:
        """Return tensor shape of this variable It is modifiable when current
        variable is not a paramter."""
        if self.value is not None:
            if isinstance(self.value, torch.Tensor):
                return list(self.value.shape)
            else:
                return self._shape
        return self._shape

    @shape.setter
    def shape(self, shape: List[Union[Text, int, None]]):
        if isinstance(shape, torch.Tensor):
            shape = shape.tolist()
        if isinstance(shape, torch.Size):
            shape = list(shape)
        if isinstance(shape, list):
            for element in shape:
                if type(element) not in {str, int}:
                    raise TypeError(
                        f'Shape of a variable should only contains int or str. however {type(element)} was given.'
                    )
        self._shape = shape

    @property
    def dtype(self) -> DataType:
        """Return tensor shape of this variable It is modifiable when current
        variable is not a paramter."""
        if self.value is not None:
            if isinstance(self.value, torch.Tensor):
                return DataType.convert_from_torch(self.value.dtype)
            if isinstance(self.value, np.ndarray):
                return DataType.convert_from_numpy(self.value.dtype)
        return self._dtype

    @dtype.setter
    def dtype(self, T: DataType):
        if isinstance(T, np.dtype):
            self._dtype = DataType.convert_from_numpy(T)
        elif isinstance(T, torch.dtype):
            self._dtype = DataType.convert_from_torch(T)
        elif isinstance(T, DataType):
            self._dtype = T
        else:
            raise TypeError(f'Invalid Dtype: {T} was given.')

    def copy(self, copy_value: bool = False):
        if not copy_value or self.value is None:
            cloned = Variable(
                name=self.name, value=self.value, is_parameter=self.is_parameter, shape=self.shape, dtype=self.dtype
            )
            return cloned

        if not isinstance(self.value, torch.Tensor):
            ppq_warning(
                f'You are requiring to copy variable {self.name}, '
                'however its value is not an instance of torch.Tensor, '
                'ppq will automaticall convert it to torch.Tensor now.'
            )
            self.value = convert_any_to_torch_tensor(self.value)
        if isinstance(self.value, torch.Tensor):
            value = self.value.clone()
        return Variable(name=self.name, value=value, is_parameter=self.is_parameter, shape=self.shape, dtype=self.dtype)


class Operation(OperationBase, Serializable):
    def __init__(
        self,
        name: str,
        op_type: str,
        attributes: Dict[str, Any],
        platform: TargetPlatform = TargetPlatform.UNSPECIFIED,
        inputs: List[Variable] = None,
        outputs: List[Variable] = None,
        opset: Opset = None,
    ) -> None:
        OperationBase.__init__(self, name, op_type, attributes, platform=platform, opset=opset)
        Serializable.__init__(self)
        self._input_vars = [] if inputs is None else inputs
        self._output_vars = [] if outputs is None else outputs

    @property
    def socket(self) -> OpSocket:
        if self.type in DEFAULT_SOCKET_TABLE:
            return DEFAULT_SOCKET_TABLE[self.type](self)
        else:
            return DEFAULT_SOCKET_CREATOR(self)

    @property
    def inputs(self) -> List[Variable]:
        return self._input_vars

    @property
    def outputs(self) -> List[Variable]:
        return self._output_vars

    @property
    def parameters(self) -> List[Variable]:
        return [var for var in self.inputs if var.is_parameter]

    @property
    def is_linear_activation(self) -> bool:
        return self.type in LINEAR_ACTIVATIONS

    @property
    def num_of_parameter(self) -> int:
        return len([var for var in self.inputs if var.is_parameter])

    @property
    def is_boundary(self) -> bool:
        up_ops, down_ops = [], []
        for var in self.inputs:
            up_ops.append(var.source_op)
        for var in self.outputs:
            down_ops.extend(var.dest_ops)
        return all([op is None for op in up_ops]) or len(down_ops) == 0

    def __hash__(self) -> int:
        return self._name.__hash__()

    def __str__(self) -> str:
        return (
            f'{self._name}(Type: {self.type}, Num of Input: {self.num_of_input}, Num of Output: {self.num_of_output})'
        )

    def __repr__(self) -> str:
        return (
            f'{self._name}(Type: {self.type}, Num of Input: {self.num_of_input}, Num of Output: {self.num_of_output})'
        )

    def __getstate__(self) -> dict:
        state = super().__getstate__()
        state['_input_vars'] = [var.name for var in self.inputs]
        state['_output_vars'] = [var.name for var in self.outputs]
        return state

    def copy(self):
        clone = Operation(
            name=self.name,
            op_type=self.type,
            attributes=self.attributes.copy(),
            platform=self.platform,
            opset=self.opset,
        )
        clone._detail = self._detail.copy()
        return clone


class BaseGraph(Serializable):
    """Graph is a PPQ Internal Represtation Data Structure.

    A computational graph is a directed graph where the nodes correspond to operations or variables.
    Variables can feed their value into operations, and operations can feed their output into other operations.
        This way, every node in the graph defines a function of the variables.

    The values that are fed into the nodes and come out of the nodes are called tensors,
        which is just a fancy word for a multi-dimensional array.
    Hence, it subsumes scalars, vectors and matrices as well as tensors of a higher rank.

    The computational graph created by PPQ contains quantization info as well as operations and variables.
        So to say it is a computational graph designed for quantization.

    All quantization related infos are stored within graph and its operations.
        See esp_ppq.IR.quantize for more information.

    Args:
        Serializable ([type]): [description]
    """

    def __init__(self, name: str, built_from: NetworkFramework = NetworkFramework.NATIVE) -> None:
        super().__init__()
        self._operations = {}
        self._variables = {}
        self._graph_inputs = {}
        self._graph_outputs = {}
        self._name = name
        self._built_from = built_from
        self._detail = {}
        self._num_of_generated_var = 0
        self._num_of_generated_op = 0

    @property
    def operations(self) -> Dict[str, Operation]:
        return self._operations

    @property
    def variables(self) -> Dict[str, Variable]:
        return self._variables

    @property
    def inputs(self) -> Dict[str, Variable]:
        return self._graph_inputs

    @property
    def outputs(self) -> Dict[str, Variable]:
        return self._graph_outputs

    def parameters(self) -> List[torch.Tensor]:
        parameters = []
        for var in self.variables.values():
            if var.is_parameter:
                parameters.append(var.value)
        return parameters

    def set_extension_attrib(self, attrib: str, value: Any):
        self._detail[attrib] = value

    @property
    def extension_attrib(self):
        return self._detail

    def append_operation(self, operation: Operation):
        if not isinstance(operation, Operation):
            raise TypeError(f'You can only insert operations via this function, however {type(operation)} was given.')
        if not all([var.name in self.variables for var in operation.inputs + operation.outputs]):
            raise KeyError(
                f'Inserting Operation {operation} has a related variable '
                'which are not included in this graph yet, insert such variables before inserting this.'
            )
        if operation.name in self.operations:
            raise KeyError(f'Duplicated Operation({operation}) was found, rename your Operation before inserting.')
        self.operations[operation.name] = operation

    def append_variable(self, var: Variable):
        if not isinstance(var, Variable):
            raise TypeError(f'You can only insert variable via this function, however {type(var)} was given.')
        if not all([dest_op.name in self.operations for dest_op in var.dest_ops]):
            raise KeyError(
                f'Inserting Variable {var} has a related Operation(dest_op) '
                'which are not included in this graph yet, insert such Operations before inserting this.'
            )
        if var.name in self.variables:
            raise KeyError(f'Duplicated Variable({var}) was found, rename your Variable before inserting.')
        self.variables[var.name] = var

    def get_downstream_operations(self, operation: Operation) -> List[Operation]:
        if not isinstance(operation, Operation):
            raise TypeError(f'Expect an operation instance, however {type(operation)} is given.')
        if operation.name not in self.operations:
            raise KeyError(f'Operation {operation.name} not in current graph.')
        downstream_ops = []
        for output_var in operation.outputs:
            downstream_ops.extend(output_var.dest_ops)
        return downstream_ops

    def get_upstream_operations(self, operation: Operation) -> List[Operation]:
        if not isinstance(operation, Operation):
            raise TypeError(f'Expect an operation instance, however {type(operation)} is given.')
        if operation.name not in self.operations:
            raise KeyError(f'Operation {operation.name} not in current graph.')
        upstream_ops = []
        for input_var in operation.inputs:
            if input_var.source_op is not None:
                upstream_ops.append(input_var.source_op)
        return upstream_ops

    def topological_sort(self) -> List[Operation]:
        visited = {operation.name: False for operation in self.operations.values()}
        sort_ret, pop_list = [], deque()
        num_of_inputs = {
            operation.name: len(self.get_upstream_operations(operation)) for operation in self.operations.values()
        }

        # initialization
        for op_name, n_input in num_of_inputs.items():
            if n_input == 0:
                pop_list.append(op_name)

        # topological sort
        for _ in range(len(visited)):
            if len(pop_list) == 0:
                break
            op_name = pop_list.popleft()
            op = self.operations[op_name]
            for post_op in self.get_downstream_operations(op):
                num_of_inputs[post_op.name] -= 1
                if num_of_inputs[post_op.name] == 0:
                    pop_list.append(post_op.name)
            visited[op.name] = True
            sort_ret.append(op)
        if all(visited.values()):
            return sort_ret
        else:
            raise RuntimeError(
                'Topological Sort failed. Some operation can not be sorted (might due to circular reference).\n'
                + ''.join(op_name + '\n' for op_name in visited if visited[op_name] == False)
            )

    def insert_op_on_var(self, inserting_op: Operation, var: str):
        """Insert one operation to current graph. Inserting operation will
        replace var.dest_ops and automatically connect to inserting_op.

        Before insertion:
            op1 -> var -> op2

        After insertion:
            op1 -> var -> inserting_op -> link_var(generated) -> op2

        ATTENTION: Inserting operation must be an empty operation with no input and output variables linked to it.

        Args:
            inserting_op (Operation): [description]
            upstream_op (Operation): [description]
            downstream_op (Operation): [description]
        """
        if not isinstance(var, str):
            raise TypeError(f'Needs a variable name(str) here, however {type(var)} was given')
        if var not in self.variables:
            raise KeyError(f'Can not inserting operation at variable {var}, variable not found.')
        if len(inserting_op.inputs) != 0 or len(inserting_op.outputs) != 0:
            raise PermissionError('Can only insert operation with no input and output variables.')

        variable = self.variables[var]

        # add to graph.
        if inserting_op.name not in self.operations.keys():
            self.append_operation(inserting_op)

        # create all links.
        name = f'PPQ_Variable_{self._num_of_generated_var}'
        self._num_of_generated_var += 1
        link_var = Variable(
            name=name, value=None, is_parameter=False, dest_ops=variable.dest_ops.copy(), source_op=inserting_op
        )
        self.append_variable(link_var)

        inserting_op.inputs.append(variable)
        inserting_op.outputs.append(link_var)

        variable.dest_ops.clear()
        variable.dest_ops.append(inserting_op)

        for op in link_var.dest_ops:
            op.inputs[op.inputs.index(variable)] = link_var

        if var in self.outputs:
            self.outputs.pop(var)
            self.outputs[link_var.name] = link_var

    def insert_op_between(self, inserting_op: Operation, up_op: Operation, down_op: Operation):
        raise Exception('This Function is removed from PPQ Since 0.6.6')

    def insert_op_before(self, A: Operation, B: Operation, input_idx: int = 0):
        """Insert an op just before given op. This function will insert given
        op A to variable B.inputs[input_idx]

        Args:
            A (Operation): Inserting Op, should has no input and output variable that links to it.
            B (Operation): before this op.
            input_idx (int, optional): For case that B has more than 1 input variable,
                user should use parameter input_idx to identify which variable is used.
        """
        if input_idx >= B.num_of_input:
            raise ValueError('Input index out of range.')
        if A.num_of_input != 0 or A.num_of_output != 0:
            raise ValueError('Can only insert op that has no input and output variable.')
        var = B.inputs[input_idx]

        var.dest_ops[var.dest_ops.index(B)] = A
        B.inputs[input_idx] = self.create_variable()
        B.inputs[input_idx].source_op = A
        B.inputs[input_idx].dest_ops.append(B)

        A.inputs.append(var)
        A.outputs.append(B.inputs[input_idx])

    def insert_op_after(self, A: Operation, B: Operation, output_idx: int = 0):
        """Insert an op just after given op. This function will insert given op
        A to variable B.outputs[output_idx]

        Args:
            A (Operation): Inserting Op, should has no input and output variable that links to it.
            B (Operation): after this op.
            output_idx (int, optional): For case that B has more than 1 output variable,
                user should use parameter output_idx to identify which variable is used.
        """
        if output_idx >= B.num_of_output:
            raise ValueError('Output index out of range.')
        if A.num_of_input != 0 or A.num_of_output != 0:
            raise ValueError('Can only insert op that has no input and output variable.')
        var = B.outputs[output_idx]

        var.source_op = A
        B.outputs[output_idx] = self.create_variable()
        B.outputs[output_idx].source_op = B
        B.outputs[output_idx].dest_ops.append(A)

        A.outputs.append(var)
        A.inputs.append(B.outputs[output_idx])

    def insert_op_between_var_and_op(self, inserting_op: Operation, up_var: Variable, down_op: Operation):
        """Insert one operation to current graph. Inserting operation will just
        between up_var and down_op.

        ATTENTION: Inserting operation must be an empty operation with no input and output variables linked to it.

        Args:
            inserting_op (Operation): [description]
            up_op (Operation): [description]
            down_op (Operation): [description]
        """
        if up_var.name not in self.variables:
            raise KeyError(f'Can not inserting operation behind {up_var.name}, variable not found.')
        if down_op.name not in self.operations:
            raise KeyError(f'Can not inserting operation behind {down_op.name}, operation not found.')
        if down_op.name not in [op.name for op in up_var.dest_ops]:
            raise PermissionError(
                f'variable {up_var.name} and {down_op.name} are not linked,'
                ' there is no way to insert an op between them.'
            )
        if len(inserting_op.inputs) != 0 or len(inserting_op.outputs) != 0:
            raise PermissionError('Can only insert operation with no input and output variables.')

        variables = []
        for var in down_op.inputs:
            if var == up_var:
                variables.append(var)
        assert len(variables) == 1, (
            f'Can not insert operation between {var.name} and {down_op.name}, graph is too complex.'
        )

        # add to graph.
        if inserting_op.name not in self.operations.keys():
            self.append_operation(inserting_op)

        # create all links.
        link_var = self.create_variable(
            name=None, value=None, is_parameter=False, dest_ops=[down_op], source_op=inserting_op
        )

        inserting_op.inputs.append(up_var)
        inserting_op.outputs.append(link_var)

        up_var.dest_ops[up_var.dest_ops.index(down_op)] = inserting_op
        down_op.inputs[down_op.inputs.index(up_var)] = link_var

    def create_link_with_op(self, A: Operation, B: Operation, variable: Variable = None):
        """Create a link with given variable from upstream_op to downstream_op
        variable will be appended to upstream_op's output and downstream_op's
        input given variable must have empty source_op or its source_op ==
        upstream_op.

        Sometime you may want to link a single upstream_op to many downstream_ops with a same variable,
            you are supposed to invoke this function for each downstream_op then.

        You can set upstream_op = None if your variable is a parameter variable.

        Example:
            create_link_with_op(var1, op1, op2)
            create_link_with_op(var1, op1, op3)

        Will makes:
                  --> op2
            op1 --|
                  --> op3

        Args:
            link_variable (Variable): _description_
            upstream_op (Operation): _description_
            downstream_op (Operation): _description_
        """
        if variable is None:
            variable = self.create_variable()
        if variable.name not in self.variables:
            raise KeyError(f'Can not find your variable {variable.name} in current graph.')
        if A is not None and A.name not in self.operations:
            raise KeyError(f'Can not find your operation {A.name} in current graph.')
        if B is not None and B.name not in self.operations:
            raise KeyError(f'Can not find your operation {B.name} in current graph.')

        if variable.source_op is None:
            variable.source_op = A
        if variable.source_op != A:
            raise PermissionError(f'Can not create link with variable {variable}, cause its source operations != {A}')

        # For complex graph, following logic might have some error.
        if A is not None and variable not in A.outputs:
            A.outputs.append(variable)
        if B is None:
            return
        if B is not None and variable not in B.inputs:
            variable.dest_ops.append(B)
            B.inputs.append(variable)
        else:
            variable.dest_ops.append(B)
            B.inputs.append(variable)
            ppq_warning(
                f'You are trying to link variable with operation, '
                f'however Variable {variable.name} has already linked with downstream op {B.name}'
            )

    def create_link_with_var(self, A: Variable, B: Variable):
        """connect upstream_variable.source_op with
        downstream_variable.dest_ops, downstream variable will be eliminated by
        this function.

        downstream_variable must have None as its source_op.

        Args:
            upstream_variable (_type_): _description_
            downstream_variable (_type_): _description_
        """
        if A is not None and A.name not in self.variables:
            raise KeyError(f'Can not find your variable {A.name} in current graph.')
        if B is not None and B.name not in self.variables:
            raise KeyError(f'Can not find your variable {B.name} in current graph.')

        if B.source_op is not None:
            raise PermissionError(
                f'Can not create link with variable {A.name} & {B.name}, '
                'Cause downstream variable has a non-empty source op'
            )

        dest_ops = B.dest_ops
        for dest_op in dest_ops:
            dest_op.inputs[dest_op.inputs.index(B)] = A
            A.dest_ops.append(dest_op)
        B.dest_ops.clear()
        self.remove_variable(B)
        return self

    def remove_operation(
        self, removing_op: Operation, keep_coherence: bool = False, remove_unlinked_variable: bool = False
    ):
        """Remove operation from graph, this function will unlink removing
        operation from current graph, pop it from graph.operations, and remove
        it from all its input and output variables.

        Parameters of this removing operations will be removed from graph by this function, without warning.

        Args:
            removing_op (Operation): [description]

            keep_coherence (bool): if keep_coherence = True,
                PPQ will link downstream operations of removing op to the upstream operation.
                if there is more than 1 input and output variable, ppq will link input[0] with output[0]
        """
        if removing_op.name not in self.operations:
            raise KeyError(f'Can not remove operation {removing_op.name}, operation not found.')

        # removing all parameters first.
        for parameter in removing_op.inputs.copy():
            if keep_coherence and removing_op.type in {'Constant', 'Identity'}:
                break
            if parameter.is_parameter:
                parameter.dest_ops.clear()
                parameter.value = None  # clear memory.
                removing_op.inputs.remove(parameter)

                self.variables.pop(parameter.name)

        related_vars = [var for var in removing_op.inputs + removing_op.outputs]
        input_var, output_var = (
            removing_op.inputs[0] if removing_op.num_of_input >= 1 else None,
            removing_op.outputs[0] if removing_op.num_of_output >= 1 else None,
        )

        # remove operation from its output variables
        for _output_var in removing_op.outputs:
            _output_var.source_op = None
        removing_op.outputs.clear()

        # remove operation from its input variables
        for _input_var in removing_op.inputs:
            if removing_op in _input_var.dest_ops:
                _input_var.dest_ops.remove(removing_op)
        removing_op.inputs.clear()

        if input_var is not None and output_var is not None and keep_coherence:
            removing_var = output_var
            dest_ops = removing_var.dest_ops
            is_graph_output = removing_var.name in self.outputs

            for op in dest_ops:
                op.inputs[op.inputs.index(removing_var)] = input_var
                input_var.dest_ops.append(op)
            removing_var.dest_ops.clear()
            removing_var.source_op = None
            self.remove_variable(removing_var)

            if is_graph_output:
                self.mark_variable_as_graph_output(input_var)

        self.operations.pop(removing_op.name)

        if remove_unlinked_variable:
            for var in related_vars:
                if var.source_op is None and len(var.dest_ops) == 0 and var.name in self.variables:
                    self.remove_variable(var)

        return self

    def remove_variable(self, removing_var: Variable):
        """Remove variable from graph, this function will unlink removing
        variable from current graph, pop it from graph.variables, and remove it
        from its source op and dest ops.

        Args:
            removing_var (Variable): [description]
        """
        if removing_var.name not in self.variables:
            raise KeyError(f'Can not remove variable {removing_var.name}, variable not found.')

        # remove from source operation
        source_op = removing_var.source_op
        if source_op is not None:
            assert isinstance(source_op, Operation), (
                f'Can not remove variable {removing_var.name}, it links to a unexpected source operation.'
            )
            if removing_var in source_op.outputs:
                source_op.outputs.remove(removing_var)
            removing_var.source_op = None

        # remove from all dest ops
        for dest_op in removing_var.dest_ops:
            assert isinstance(dest_op, Operation), (
                f'Can not remove variable {removing_var.name}, it links to a unexpected dest operation.'
            )
            if removing_var in dest_op.inputs:
                dest_op.inputs.remove(removing_var)
        removing_var.dest_ops.clear()

        if removing_var.name in self.outputs:
            self.outputs.pop(removing_var.name)

        if removing_var.name in self.inputs:
            self.inputs.pop(removing_var.name)

        self.variables.pop(removing_var.name)
        return self

    def create_operation(
        self,
        op_type: str,
        name: str = None,
        attributes: Dict[str, Any] = None,
        platform: TargetPlatform = TargetPlatform.UNSPECIFIED,
        inputs: List[Variable] = None,
        outputs: List[Variable] = None,
        **kwargs,
    ) -> Operation:
        """Create an operation and attach it it current graph. op_type is
        mandatory here, however op_name is not required. PPQ will automatically
        generates a name for your operation:
        PPQ_Operation_{self._num_of_generated_op}.

        Use this function carefully, cause once your network is quantized,
            simply create an operation via this function might cause unexpected error.
        Beawre that operation created by this function has no meta data and quantization info,
            which is needed to export and executing your graph.

        Do not set inputs and outputs via this function,
            to link your operation with others, use graph.create_link_with_var instead.

        Args:
            op_type (str): _description_
            name (str, optional): _description_. Defaults to None.
            attributes (Dict[str, Any], optional): _description_. Defaults to None.
            platform (TargetPlatform, optional): _description_. Defaults to TargetPlatform.UNSPECIFIED.
            inputs (List[Variable], optional): _description_. Defaults to None.
            outputs (List[Variable], optional): _description_. Defaults to None.

        Returns:
            Operation: _description_
        """
        if inputs is None:
            inputs = []
        if outputs is None:
            outputs = []

        if name is None:
            name = f'PPQ_Operation_{self._num_of_generated_op}'
            self._num_of_generated_op += 1

        if not isinstance(inputs, list):
            raise TypeError(
                f'A list of input variable is required for creating operation, however {type(inputs)} was given'
            )

        if attributes is None:
            attributes = {}
        created = Operation(
            name=name, op_type=op_type, attributes=attributes, platform=platform, inputs=inputs, outputs=outputs
        )
        self.append_operation(created)

        for item in inputs:
            if not isinstance(item, Variable):
                raise TypeError(
                    f'A list contains variables is required for creating operation, '
                    f'however there is a {type(item)} in your input list.'
                )
            item.dest_ops.append(created)

        if not isinstance(outputs, list):
            raise TypeError(
                f'A list of output variable is required for creating operation, however {type(inputs)} was given'
            )
        for item in outputs:
            if not isinstance(item, Variable):
                raise TypeError(
                    f'A list contains variables is required for creating operation, '
                    f'however there is a {type(item)} in your output list.'
                )
            item.source_op = created
        return created

    def create_variable(
        self,
        name: str = None,
        value: Any = None,
        is_parameter: bool = False,
        dest_ops: List[OperationBase] = None,
        source_op: OperationBase = None,
        **kwargs,
    ) -> Variable:
        """Create a variable and attach it it current graph. PPQ will
        automatically generates a name for your variable:
        PPQ_Variable_{self._num_of_generated_op}.

        Use this function carefully, cause once your network is quantized,
            simply create an variable via this function might cause unexpected error.
        You'd better invoke this function before running your quantizer.

        If dest_ops and source_op is not None, this function will auto link created variable with them.

        Args:
            name (str, optional): _description_. Defaults to None.
            value (Any, optional): _description_. Defaults to None.
            is_parameter (bool, optional): _description_. Defaults to False.
            dest_ops (List[OperationBase], optional): _description_. Defaults to None.
            source_op (OperationBase, optional): _description_. Defaults to None.

        Returns:
            Variable: _description_
        """
        if name is None:
            name = f'PPQ_Variable_{self._num_of_generated_var}'
            self._num_of_generated_var += 1

        created = Variable(name=name, value=value, is_parameter=is_parameter, dest_ops=dest_ops, source_op=source_op)

        if dest_ops is not None:
            if not isinstance(dest_ops, list):
                raise TypeError(
                    f'Parameter dest ops should be a list of Operation, however {type(dest_ops)} was given.'
                )
            for op in dest_ops:
                if not isinstance(op, Operation):
                    raise TypeError(f'Parameter dest ops should be a list of Operation, however {type(op)} was given.')
                op.inputs.append(created)

        if source_op is not None:
            if not isinstance(source_op, Operation):
                raise TypeError(f'Parameter dest ops should be an Operation, however {type(source_op)} was given.')
            source_op.outputs.append(created)

        self.append_variable(created)
        return created

    def mark_variable_as_graph_input(self, var: Variable):
        if not isinstance(var, Variable):
            raise TypeError(f'Except a variable here, however {type(var)} was given.')
        var_name = var.name
        if var_name not in self.variables:
            raise KeyError(f'Can not find variable {var_name} within current graph.')
        if var_name in self.inputs:
            return
        if var_name in self.outputs:
            raise KeyError(f'Can not mark variable {var_name} as graph input, cause it is graph output.')
        self.inputs[var_name] = self.variables[var_name]

    def mark_variable_as_graph_output(self, var: Variable):
        if not isinstance(var, Variable):
            raise TypeError(f'Except a variable here, however {type(var)} was given.')
        var_name = var.name
        if var_name not in self.variables:
            raise KeyError(f'Can not find variable {var_name} within current graph.')
        if var_name in self.outputs:
            return
        self.outputs[var_name] = self.variables[var_name]

    def __getstate__(self) -> dict:
        state = super().__getstate__()
        state['_graph_inputs'] = [var for var in self.inputs]
        state['_graph_outputs'] = [var for var in self.outputs]
        return state

    def copy(self, copy_value: bool = False):
        """Clone current graph. Use parameter copy_value to control whether to
        do a Shallow Copy or Deep Copy.

        For copy_value = True, there will be a copy of each parameter in your network.
            ATTENTION: it might cause gpu memory overflow.
        For copy_value = False, cloned network will share the same parameter tensor of current one.

        ATTENTION: all quantization config will be cloned,
            all scales and offsets will be cloned even with copy_valye = False.

        Shallow Copy: Shallow repetition is quicker.
        However, it's “lazy” it handles pointers and references.
        Rather than creating a contemporary copy of the particular knowledge the pointer points to,
            it simply copies over the pointer price.
        So, each the first and therefore the copy can have pointers that reference constant underlying knowledge.

        Deep Copy: Deep repetition truly clones the underlying data.
        It is not shared between the first and therefore the copy.
        """
        from esp_ppq.IR.quantize import QuantableOperation

        cloned = BaseGraph(name=self._name, built_from=self._built_from)
        for op in self.operations.values():
            cloned.append_operation(op.copy())
        for var in self.variables.values():
            cloned.append_variable(var.copy(copy_value=copy_value))

        # notice that all operations is copyed without link, so do all variables
        # relink them with following code
        config_dict = {}
        for op in self.operations.values():
            assert op.name in cloned.operations, f'Graph Copy Error, Operation {op.name} is not correctly cloned'
            c_op = cloned.operations[op.name]
            for i_var in op.inputs:
                assert i_var.name in cloned.variables, (
                    f'Graph Copy Error, Variable {i_var.name} is not correctly cloned'
                )
                ci_var = cloned.variables[i_var.name]
                cloned.create_link_with_op(variable=ci_var, A=ci_var.source_op, B=c_op)
            for o_var in op.outputs:
                assert o_var.name in cloned.variables, (
                    f'Graph Copy Error, Variable {o_var.name} is not correctly cloned'
                )
                co_var = cloned.variables[o_var.name]
                c_op.outputs.append(co_var)
                co_var.source_op = c_op
            if isinstance(op, QuantableOperation):
                for cfg, var in op.config_with_variable:
                    config_dict[cfg._hash] = (op, var)

        # relink config to there cloned master.
        for c_op in cloned.operations.values():
            if isinstance(c_op, QuantableOperation):
                for cfg, var in c_op.config_with_variable:
                    if cfg.dominated_by != cfg:
                        assert cfg.dominated_by._hash in config_dict, (
                            'Graph Copy Error, can not find a corresponding master config.'
                        )
                        op, var = config_dict[cfg.dominated_by._hash]

                        op = cloned.operations[op.name]
                        assert isinstance(op, QuantableOperation), 'Graph Copy Error, Unexpected Master Operation Type.'
                        for mcfg, mvar in op.config_with_variable:
                            if mvar.name == var.name:
                                cfg._dominator = mcfg

        # recreate input, output
        for name in self.inputs:
            cloned.inputs[name] = cloned.variables[name]
        for name in self.outputs:
            cloned.outputs[name] = cloned.variables[name]

        # check integrity
        for op in self.operations.values():
            if op.name not in cloned.operations:
                raise KeyError(f'Graph Copy Error, Operation {op.name} is Missing')
        for var in self.variables.values():
            if var.name not in cloned.variables:
                raise KeyError(f'Graph Copy Error, Variable {var.name} is Missing')
        for name in self.inputs:
            if name not in cloned.inputs:
                raise KeyError(f'Graph Copy Error, Input {var.name} is Missing')
        for name in self.outputs:
            if name not in cloned.outputs:
                raise KeyError(f'Graph Copy Error, Output {var.name} is Missing')
        cloned._num_of_generated_op = self._num_of_generated_op
        cloned._num_of_generated_var = self._num_of_generated_var
        cloned._detail = self._detail.copy()
        return cloned


class GraphBuilder(metaclass=SingletonMeta):
    @abstractmethod
    def build(self, file_path: str, **kwargs) -> BaseGraph:
        pass


class GraphExporter(metaclass=SingletonMeta):
    @abstractmethod
    def export(self, file_path: str, graph: BaseGraph, config_path: str = None, **kwargs):
        pass


class OperationExporter(metaclass=SingletonMeta):
    @abstractmethod
    def export(self, operation: Operation, graph: BaseGraph, **kwargs) -> Operation:
        pass



import torch
import torchvision

from esp_ppq import *
from esp_ppq.api import *


# ------------------------------------------------------------
# 在这个脚本中，我们将向你展示如何自由调度算子，并实现混合精度推理
# 在开始之前，我们首先设计一个支持混合精度的量化器
# ------------------------------------------------------------
class MyQuantizer(BaseQuantizer):
    # ------------------------------------------------------------
    # quant_operation_types 是一个类型枚举，在这里你需要写下所有该量化器所需要量化的算子
    # ------------------------------------------------------------
    @property
    def quant_operation_types(self) -> set:
        return {'Conv'}

    # ------------------------------------------------------------
    # 一旦你确定了那些算子需要量化，则需要在 init_quantize_config 为他们初始化量化信息
    # 然而你需要注意的是，由于手动调度的存在，用户可以强制调度一个类型不在 quant_operation_types 中的算子来到量化平台
    # 我建议你针对这类情况进行回应。或者，在探测到算子类型并非可量化类型后进行报错
    # ------------------------------------------------------------
    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        # ------------------------------------------------------------
        # 为卷积算子初始化量化信息，只量化卷积算子的输入(input & weight)，bias 不做量化
        # ------------------------------------------------------------
        if operation.type == 'Conv':
            config = self.create_default_quant_config(
                op=operation,
                num_of_bits=4,
                quant_max=15,
                quant_min=-16,
                observer_algorithm='percentile',
                policy=QuantizationPolicy(
                    QuantizationProperty.PER_TENSOR + QuantizationProperty.LINEAR + QuantizationProperty.SYMMETRICAL
                ),
                rounding=RoundingPolicy.ROUND_HALF_EVEN,
            )

            # ------------------------------------------------------------
            # 关闭所有输出量化，状态设置为fp32
            # ------------------------------------------------------------
            for tensor_quant_config in config.output_quantization_config:
                tensor_quant_config.state = QuantizationStates.FP32

            # ------------------------------------------------------------
            # 关闭 bias 量化，状态设置为fp32
            # ------------------------------------------------------------
            if operation.num_of_input == 3:
                config.input_quantization_config[-1].state = QuantizationStates.FP32

            # ------------------------------------------------------------
            # 如果算子被调度到 INT8 平台上，执行 INT8 的量化
            # ------------------------------------------------------------
            if operation.platform == TargetPlatform.ACADEMIC_INT8:
                print(f'{operation.name} has been dispatched to INT8')
                config.input_quantization_config[0].num_of_bits = 8
                config.input_quantization_config[0].quant_max = 127
                config.input_quantization_config[0].quant_min = -128

                config.input_quantization_config[1].num_of_bits = 8
                config.input_quantization_config[1].quant_max = 127
                config.input_quantization_config[1].quant_min = -128

            return config
        else:
            raise TypeError(f'Unsupported Op Type: {operation.type}')

    # ------------------------------------------------------------
    # 当前量化器进行量化的算子都将被发往一个指定的目标平台
    # 这里我们选择 TargetPlatform.ACADEMIC_INT4 作为目标平台
    # ------------------------------------------------------------
    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ACADEMIC_INT4


# 注册我们的量化器
register_network_quantizer(MyQuantizer, platform=TargetPlatform.ACADEMIC_INT4)

# ------------------------------------------------------------
# 下面，我们向你展示 PPQ 的手动调度逻辑
# 我们仍然以 MobilenetV2 举例，向你展示如何完成混合精度调度
# ------------------------------------------------------------

BATCHSIZE = 32
INPUT_SHAPE = [BATCHSIZE, 3, 224, 224]
DEVICE = 'cuda'
PLATFORM = TargetPlatform.ACADEMIC_INT4
CALIBRATION = [torch.rand(size=INPUT_SHAPE) for _ in range(32)]


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# ------------------------------------------------------------
# 为了执行手动调度，你必须首先创建 QuantizationSetting
# 使用 QuantizationSetting.dispatch_table 属性来传递调度方案
# 大多数预制量化器没有针对 INT4 写过量化方案，因此只支持 FP32 - INT8 的相互调度
# ------------------------------------------------------------
QS = QuantizationSettingFactory.default_setting()

# 示例语句：下面的语句将把 Op1 调度到 FP32 平台
# QS.dispatching_table.append(operation='Op1', platform=TargetPlatform.FP32)

# ------------------------------------------------------------
# 我如何知道调度那些层到高精度会得到最优的性能表现？
# layerwise_error_analyse 函数正是为此设计的，我们调用该方法
# 然后选择那些误差较高的层调度到高精度平台

# 我们首先将所有层全部送往 INT4，然后执行误差分析
# ------------------------------------------------------------

# ------------------------------------------------------------
# 如果你使用 ENABLE_CUDA_KERNEL 方法
# PPQ 将会尝试编译自定义的高性能量化算子，这一过程需要编译环境的支持
# 如果你在编译过程中发生错误，你可以删除此处对于 ENABLE_CUDA_KERNEL 方法的调用
# 这将显著降低 PPQ 的运算速度；但即使你无法编译这些算子，你仍然可以使用 pytorch 的 gpu 算子完成量化
# ------------------------------------------------------------
with ENABLE_CUDA_KERNEL():
    dump_torch_to_onnx(
        model=model, onnx_export_file='Output/model.onnx', input_shape=INPUT_SHAPE, input_dtype=torch.float32
    )
    graph = load_onnx_graph(onnx_import_file='Output/model.onnx')
    quantized = quantize_native_model(
        model=graph,
        calib_dataloader=CALIBRATION,
        calib_steps=32,
        input_shape=INPUT_SHAPE,
        collate_fn=collate_fn,
        platform=PLATFORM,
        device=DEVICE,
        verbose=0,
        setting=QS,
    )

    # ------------------------------------------------------------
    # 使用 graphwise_analyse 衡量调度前的量化误差
    # ------------------------------------------------------------
    reports = graphwise_error_analyse(
        graph=quantized, running_device=DEVICE, collate_fn=collate_fn, dataloader=CALIBRATION
    )

    # ------------------------------------------------------------
    # 执行逐层分析，结果是一个字典，该字典内写入了所有算子的单层量化误差
    # ------------------------------------------------------------
    reports = layerwise_error_analyse(
        graph=quantized, running_device=DEVICE, collate_fn=collate_fn, dataloader=CALIBRATION, verbose=False
    )

    # 从大到小排序单层误差
    sensitivity = [(op_name, error) for op_name, error in reports.items()]
    sensitivity = sorted(sensitivity, key=lambda x: x[1], reverse=True)

    # ------------------------------------------------------------
    # 将前十个误差最大的层送上INT8，并重新量化
    # ------------------------------------------------------------
    for op_name, _ in sensitivity[:10]:
        QS.dispatching_table.append(operation=op_name, platform=TargetPlatform.ACADEMIC_INT8)
    graph = load_onnx_graph(onnx_import_file='Output/model.onnx')
    quantized = quantize_native_model(
        model=graph,
        calib_dataloader=CALIBRATION,
        calib_steps=32,
        input_shape=INPUT_SHAPE,
        collate_fn=collate_fn,
        platform=PLATFORM,
        device=DEVICE,
        verbose=0,
        setting=QS,
    )

    # ------------------------------------------------------------
    # 使用 graphwise_analyse 衡量最终的量化误差
    # ------------------------------------------------------------
    reports = graphwise_error_analyse(
        graph=quantized, running_device=DEVICE, collate_fn=collate_fn, dataloader=CALIBRATION
    )



from time import time

import numpy as np
import openvino.runtime as ov
from tqdm import tqdm

if ov.get_version() < '2022.1.0':
    raise Exception('Please Install Openvino >= 2022.1.0')


def Benchmark(ir_or_onnx_file: str, samples: int = 500, jobs: int = 4) -> float:
    """Run Performance Benckmark with given onnx model. (Or Openvino IR)

    By default this function will run with Async Mode.
    """
    # https://docs.openvino.ai/latest/api/ie_python_api/_autosummary/openvino.runtime.InferRequest.html
    core = ov.Core()
    # core.add_extension("path_to_extension_library.so")
    model = core.read_model(ir_or_onnx_file)
    compiled_model = core.compile_model(model, 'CPU')

    infer_request = compiled_model.create_infer_request()
    print(
        f'Openvino Model Loaded: {len(infer_request.input_tensors)} Input Tensors, {len(infer_request.output_tensors)} Output Tensors'
    )

    feed_dict = []
    for tensor in infer_request.input_tensors:
        feed_dict.append(np.random.random(size=tensor.shape).astype(tensor.element_type.to_dtype()))

    # Start async inference on a single infer request
    infer_request.start_async()
    # Wait for 1 milisecond
    infer_request.wait_for(1)
    # Wait for inference completion
    infer_request.wait()
    infer_queue = ov.AsyncInferQueue(compiled_model, jobs=jobs)

    tick = time()
    for _ in tqdm(range(samples)):
        # Wait for at least one available infer request and start asynchronous inference
        infer_queue.start_async(feed_dict)
    # Wait for all requests to complete
    infer_queue.wait_all()
    tok = time()

    print(f'Time span: {tok - tick: .4f} sec')
    return tick - tok

    """
    infer_request.infer(feed_dict)
    for record in infer_request.get_profiling_info():
        print(record.node_name, record.node_type, record.cpu_time.total_seconds(), record.real_time.total_seconds())
    """



from typing import Callable, Iterable

import torch
import torchvision

from esp_ppq import (
    BaseGraph,
    QuantizationOptimizationPass,
    QuantizationOptimizationPipeline,
    QuantizationSetting,
    TargetPlatform,
)
from esp_ppq.api import ENABLE_CUDA_KERNEL
from esp_ppq.executor.torch import TorchExecutor
from esp_ppq.IR.quantize import QuantableOperation
from esp_ppq.IR.search import SearchableGraph
from esp_ppq.quantization.optim import (
    ParameterQuantizePass,
    PassiveParameterQuantizePass,
    QuantAlignmentPass,
    QuantizeSimplifyPass,
    RuntimeCalibrationPass,
)
from esp_ppq.quantization.quantizer import TensorRTQuantizer

# ------------------------------------------------------------
# 在这个例子中，我们将向你介绍如何自定义量化优化过程，以及如何手动调用优化过程
# ------------------------------------------------------------

BATCHSIZE = 32
INPUT_SHAPE = [BATCHSIZE, 3, 224, 224]
DEVICE = 'cuda'
PLATFORM = TargetPlatform.TRT_INT8


# ------------------------------------------------------------
# 和往常一样，我们要创建 calibration 数据，以及加载模型
# ------------------------------------------------------------
def load_calibration_dataset() -> Iterable:
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]


CALIBRATION = load_calibration_dataset()


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# ------------------------------------------------------------
# 下面，我们将向你展示如何自定义图融合过程
# 图融合过程将改变量化方案，PPQ 使用 Tensor Quantization Config
# 来描述图融合的具体规则，其底层由并查集进行实现
# ------------------------------------------------------------


# ------------------------------------------------------------
# 定义我们自己的图融合过程，在这里我们将尝试进行 Conv - Clip 的融合
# 但与平常不同的是，我们将关闭 Clip 之后的量化点，保留 Conv - Clip 中间的量化
# 对于更为复杂的模式匹配，你可以参考 esp_ppq.quantization.optim.refine.SwishFusionPass
# ------------------------------------------------------------
class MyFusion(QuantizationOptimizationPass):
    def optimize(
        self, graph: BaseGraph, dataloader: Iterable, collate_fn: Callable, executor: TorchExecutor, **kwargs
    ) -> None:
        # 图融合过程往往由图模式匹配开始，让我们建立一个模式匹配引擎
        search_engine = SearchableGraph(graph=graph)
        for pattern in search_engine.pattern_matching(patterns=['Conv', 'Clip'], edges=[[0, 1]], exclusive=True):
            conv, relu = pattern

            # 匹配到图中的 conv - relu 对，接下来关闭不必要的量化点
            # 首先我们检查 conv - relu 是否都是量化算子，是否处于同一平台
            is_quantable = isinstance(conv, QuantableOperation) and isinstance(relu, QuantableOperation)
            is_same_plat = conv.platform == relu.platform

            if is_quantable and is_same_plat:
                # 将 relu 输入输出的量化全部指向 conv 输出
                # 一旦调用 dominated_by 完成赋值，则调用 dominated_by 的同时
                # PPQ 会将 relu.input_quant_config[0] 与 relu.output_quant_config[0] 的状态置为 OVERLAPPED
                # 在后续运算中，它们所对应的量化不再起作用
                relu.input_quant_config[0].dominated_by = conv.output_quant_config[0]
                relu.output_quant_config[0].dominated_by = conv.output_quant_config[0]


# ------------------------------------------------------------
# 自定义图融合的过程将会干预量化器逻辑，我们需要新建量化器
# 此处我们继承 TensorRT Quantizer，算子的量化逻辑将使用 TensorRT 的配置
# 但在生成量化管线时，我们将覆盖量化器原有的逻辑，使用我们自定义的管线
# 这样我们就可以把自定义的图融合过程放置在合适的位置上，而此时 QuantizationSetting 也不再起作用
# ------------------------------------------------------------
class MyQuantizer(TensorRTQuantizer):
    def build_quant_pipeline(self, setting: QuantizationSetting) -> QuantizationOptimizationPipeline:
        return QuantizationOptimizationPipeline(
            [
                QuantizeSimplifyPass(),
                ParameterQuantizePass(),
                MyFusion(name='My Optimization Procedure'),
                RuntimeCalibrationPass(),
                QuantAlignmentPass(),
                PassiveParameterQuantizePass(),
            ]
        )


from esp_ppq.api import quantize_torch_model, register_network_quantizer

register_network_quantizer(quantizer=MyQuantizer, platform=TargetPlatform.EXTENSION)

# ------------------------------------------------------------
# 如果你使用 ENABLE_CUDA_KERNEL 方法
# PPQ 将会尝试编译自定义的高性能量化算子，这一过程需要编译环境的支持
# 如果你在编译过程中发生错误，你可以删除此处对于 ENABLE_CUDA_KERNEL 方法的调用
# 这将显著降低 PPQ 的运算速度；但即使你无法编译这些算子，你仍然可以使用 pytorch 的 gpu 算子完成量化
# ------------------------------------------------------------
with ENABLE_CUDA_KERNEL():
    quantized = quantize_torch_model(
        model=model,
        calib_dataloader=CALIBRATION,
        calib_steps=32,
        input_shape=INPUT_SHAPE,
        collate_fn=collate_fn,
        platform=TargetPlatform.EXTENSION,
        onnx_export_file='model.onnx',
        device=DEVICE,
        verbose=0,
    )



from typing import Iterable

import torch
import torchvision

from esp_ppq import BaseQuantizer, OperationQuantizationConfig, TargetPlatform
from esp_ppq.api import (
    ENABLE_CUDA_KERNEL,
    export_ppq_graph,
    quantize_torch_model,
    register_network_exporter,
    register_network_quantizer,
)
from esp_ppq.core import (
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    convert_any_to_numpy,
)
from esp_ppq.IR import BaseGraph, Operation, QuantableOperation

# ------------------------------------------------------------
# 在这个脚本中，我们将创建一个新的量化平台，定义我们自己的量化规则
# 这意味着我们将创建自己的 Quantizer，并调节 QuantSetting 中的各项属性
# ------------------------------------------------------------

BATCHSIZE = 32
INPUT_SHAPE = [BATCHSIZE, 3, 224, 224]
DEVICE = 'cuda'
PLATFORM = TargetPlatform.TRT_INT8


def load_calibration_dataset() -> Iterable:
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]


CALIBRATION = load_calibration_dataset()


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)


# ------------------------------------------------------------
# 定义自己的量化器，并完成注册
# 自定义量化器需要你完成所有接口函数与接口属性（接口属性必须以@property修饰）
# ------------------------------------------------------------
class MyQuantizer(BaseQuantizer):
    # ------------------------------------------------------------
    # quant_operation_types 是一个类型枚举，在这里你需要写下所有该量化器所需要量化的算子
    # ------------------------------------------------------------
    @property
    def quant_operation_types(self) -> set:
        return {'Conv'}

    # ------------------------------------------------------------
    # 一旦你确定了那些算子需要量化，则需要在 init_quantize_config 为他们初始化量化信息
    # 然而你需要注意的是，由于手动调度的存在，用户可以强制调度一个类型不在 quant_operation_types 中的算子来到量化平台
    # 我建议你针对这类情况进行回应。或者，在探测到算子类型并非可量化类型后进行报错
    # ------------------------------------------------------------
    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        # ------------------------------------------------------------
        # 为卷积算子初始化量化信息，只量化卷积算子的输出
        # ------------------------------------------------------------
        if operation.type == 'Conv':
            config = self.create_default_quant_config(
                op=operation,
                num_of_bits=8,
                quant_max=127,
                quant_min=-128,
                observer_algorithm='percentile',
                policy=QuantizationPolicy(
                    QuantizationProperty.PER_TENSOR + QuantizationProperty.LINEAR + QuantizationProperty.SYMMETRICAL
                ),
                rounding=RoundingPolicy.ROUND_HALF_EVEN,
            )

            # ------------------------------------------------------------
            # 关闭所有输入量化，状态设置为fp32
            # ------------------------------------------------------------
            for tensor_quant_config in config.input_quantization_config:
                tensor_quant_config.state = QuantizationStates.FP32

            return config
        else:
            raise TypeError(f'Unsupported Op Type: {operation.type}')

    # ------------------------------------------------------------
    # 当前量化器进行量化的算子都将被发往一个指定的目标平台
    # 这里我们选择 TargetPlatform.EXTENSION 作为目标平台
    # ------------------------------------------------------------
    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.EXTENSION

    @property
    def activation_fusion_types(self) -> set:
        # 列举此处的算子会与他们之前的 卷积 和 矩阵乘法执行激活函数图融合
        # 从而消去一些错误的量化行为，对于更为复杂的图融合，你必须手写一个单独的优化过程
        # 不过我们这里没有声明 relu, clip 是量化类型，因此图融合并不会起作用

        # 图融合起作用的前提是参与融合的所有算子，全部都需要被量化
        return {'Relu', 'Clip'}


# ------------------------------------------------------------
# 注册我们的量化器，目标平台为 TargetPlatform.EXTENSION
# 后面我们将以这个平台来调用量化器
# ------------------------------------------------------------
register_network_quantizer(MyQuantizer, TargetPlatform.EXTENSION)

with ENABLE_CUDA_KERNEL():
    # ------------------------------------------------------------
    # 以 TargetPlatform.EXTENSION 作为目标平台调用量化器
    # ------------------------------------------------------------
    quantized = quantize_torch_model(
        model=model,
        calib_dataloader=CALIBRATION,
        calib_steps=32,
        input_shape=INPUT_SHAPE,
        collate_fn=collate_fn,
        platform=TargetPlatform.EXTENSION,
        onnx_export_file='Output/onnx.model',
        device=DEVICE,
        verbose=0,
    )

import onnx

# ------------------------------------------------------------
# 完成量化后，我们开始导出量化信息，这里我们注册一个网络量化信息导出器
# 对于一个网络导出器而言，你需要关注以下两件事：
#   1. 对于一些场景而言，你需要导出 PPQ 的量化信息到特定的文件
#   2. 有些时候你使用 PPQ 训练过你的网络，你可能还需要导出训练后的网络权重
# 我们这里向你展示如何导出量化信息到文本文档，而网络权重的导出则继承于 onnx
# ------------------------------------------------------------
from esp_ppq.parser import OnnxExporter


class MyExporter(OnnxExporter):
    def convert_value(self, value: torch.Tensor) -> str:
        if type(value) in {int, float}:
            return value
        else:
            value = convert_any_to_numpy(value, accept_none=True)
            if value is None:
                return value  # SOI config has Nona as its scale and
            return value.tolist()

    def export(self, file_path: str, graph: BaseGraph, config_path: str, **kwargs):
        # ------------------------------------------------------------
        # 接下来我们将导出量化信息，在 PPQ 中所有的量化信息都绑定在 Op 上
        # 因此我们需要遍历图中所有的 Op, 将绑定在其上的量化信息导出到文件
        # ------------------------------------------------------------
        with open(config_path, 'w') as file:
            for name, op in graph.operations.items():
                if not isinstance(op, QuantableOperation):
                    continue

                for cfg, var in op.config_with_variable:
                    file.write(f"{name}: {var.name}\n")
                    file.write(f"Quant State: {cfg.state.name}\n")
                    file.write(f"Scale:  {self.convert_value(cfg.scale)}\n")
                    file.write(f"Offset: {self.convert_value(cfg.offset)}\n")

        # ------------------------------------------------------------
        # 最后我们导出完整的计算图到 onnx
        # ------------------------------------------------------------
        onnx.save(self.export_graph(graph=graph), file_path)


# ------------------------------------------------------------
# 注册导出器并导出模型
# ------------------------------------------------------------
register_network_exporter(exporter=MyExporter, platform=TargetPlatform.EXTENSION)
export_ppq_graph(
    graph=quantized,
    platform=TargetPlatform.EXTENSION,
    graph_save_to='Output/model.onnx',
    config_save_to='Output/model.json',
)



from .torch import (
    ACADEMIC_BACKEND_TABLE,
    DEFAULT_BACKEND_TABLE,
    ESPDL_QUANT_BACKEND_TABLE,
    EXTENSION_BACKEND_TABLE,
    NXP_BACKEND_TABLE,
    ONNX_BACKEND_TABLE,
    PPL_DSP_BACKEND_TABLE,
    PPL_GPU_BACKEND_TABLE,
    TorchBackendContext,
)
